#!/usr/bin/env python3
#
# The Qubes OS Project, http://www.qubes-os.org
#
# Copyright 2010 Rafal Wojtczuk <rafal@invisiblethingslab.com>
# Copyright 2020 Norbert Kamiński <norbert.kaminski@3mdeb.com>
#
# SPDX-License-Identifier: LGPL-2.1-or-later
#

import hashlib
import os
import shutil
import subprocess
import tempfile
from qubes_fwupd_common import create_dirs

FWUPD_DOM0_DIR = "/var/cache/fwupd/qubes"
FWUPD_DOM0_UPDATES_DIR = os.path.join(FWUPD_DOM0_DIR, "updates")
FWUPD_DOM0_UNTRUSTED_DIR = os.path.join(FWUPD_DOM0_UPDATES_DIR, "untrusted")
FWUPD_DOM0_METADATA_DIR = os.path.join(FWUPD_DOM0_DIR, "metadata")

FWUPD_VM_DIR = "/home/user/.cache/fwupd"
FWUPD_VM_UPDATES_DIR = os.path.join(FWUPD_VM_DIR, "updates")
FWUPD_VM_METADATA_DIR = os.path.join(FWUPD_VM_DIR, "metadata")
FWUPD_PKI = "/etc/pki/fwupd"
FWUPD_DOWNLOAD_PREFIX = "https://fwupd.org/downloads/"
HEADS_UPDATES_DIR = "/boot/updates"


class FwupdReceiveUpdates:
    def _check_shasum(self, file_path, sha):
        """Compares computed SHA256 checksum with `sha` parameter.

        Keyword arguments:
        file_path -- absolute path to the file
        sha -- SHA256 checksum of the file
        """
        with open(file_path, "rb") as f:
            c_sha = hashlib.sha256(f.read()).hexdigest()
        if c_sha != sha:
            self.clean_cache()
            raise ValueError(f"Computed checksum {c_sha} did NOT match {sha}.")

    def _jcat_verification_without_postquantum(self, verification_output):
        """Return True e ven if the PQ CA verification fails as long as
        at least sha256 plus one pkcs7 entry passed for the classic LVFS CA.
        """
        passed = set()
        for line in verification_output.splitlines():
            line = line.strip()
            if line.startswith("PASSED "):
                kind = line.split()[1].rstrip(":")
                passed.add(kind)
            elif line.startswith("FAILED ") and "FAILED: " not in line:
                if "LVFS CA 2025PQ" in line and "[-89]" in line:
                    print(
                        "--ignore-postquantum is set, ignoring LVFS CA 2025PQ failure"
                    )
                    continue
                return False
        return "sha256" in passed and "pkcs7" in passed

    def _jcat_verification(self, file_path, file_directory, ignore_postquantum=False):
        """Verifies SHA1+SHA256 checksum and PKCS#7 signature.

        Keyword argument:
        file_path -- absolute path to jcat file
        file_directory -- absolute path to the directory to jcat file location
        """
        assert file_path.startswith("/"), f"bad file path {file_path!r}"
        cmd_jcat = ["jcat-tool", "verify", file_path, "--public-keys", FWUPD_PKI]
        environ = os.environ.copy()
        environ["LC_ALL"] = "C"
        p = subprocess.Popen(
            cmd_jcat,
            cwd=file_directory,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            env=environ,
        )
        stdout, stderr = p.communicate()
        verification = stdout.decode("utf-8")
        print(verification)
        if p.returncode != 0:
            if ignore_postquantum and self._jcat_verification_without_postquantum(
                verification
            ):
                return
            self.clean_cache()
            stderr_str = stderr.decode()
            if "SignatureNotValidYet" in stderr_str:
                raise Exception(
                    "PGP signature creation time is in the"
                    " future (update-vm clock may be out of sync).\n"
                    "Run /usr/bin/qvm-sync-clock to update the time.\n" + stderr_str
                )
            else:
                raise Exception("jcat-tool: Verification failed")

    def handle_fw_update(self, updatevm, sha, filename):
        """Copies firmware update archives from the updateVM.

        Keyword arguments:
        updatevm -- update VM name
        sha -- SHA256 checksum of the firmware update archive
        filename -- name of the firmware update archive
        """
        create_dirs(FWUPD_DOM0_UPDATES_DIR, FWUPD_DOM0_UNTRUSTED_DIR)

        with tempfile.TemporaryDirectory(dir=FWUPD_DOM0_UNTRUSTED_DIR) as tmpdir:
            dom0_firmware_untrusted_path = os.path.join(tmpdir, filename)
            updatevm_firmware_file_path = os.path.join(FWUPD_VM_UPDATES_DIR, filename)

            cmd_copy = [
                "qvm-run",
                "--pass-io",
                "--no-gui",
                "-q",
                "-a",
                "--no-shell",
                "--",
                updatevm,
                "cat",
                "--",
                updatevm_firmware_file_path,
            ]
            with open(dom0_firmware_untrusted_path, "bx") as untrusted_file:
                p = subprocess.Popen(cmd_copy, stdout=untrusted_file, shell=False)
            p.wait()
            if p.returncode != 0:
                raise Exception("qvm-run: Copying firmware file failed!!")

            self._check_shasum(dom0_firmware_untrusted_path, sha)
            # jcat verification will be done by fwupd itself
            self.arch_name = filename
            self.arch_path = os.path.join(FWUPD_DOM0_UPDATES_DIR, filename)
            shutil.move(dom0_firmware_untrusted_path, self.arch_path)

    def handle_metadata_update(self, updatevm, metadata_url, ignore_postquantum=False):
        """Copies metadata files from the updateVM.

        Keyword argument:
        updatevm -- update VM name
        """
        metadata_name = os.path.basename(metadata_url)
        self.metadata_file = os.path.join(FWUPD_DOM0_METADATA_DIR, metadata_name)
        self.metadata_file_jcat = self.metadata_file + ".jcat"
        self.metadata_file_updatevm = os.path.join(FWUPD_VM_METADATA_DIR, metadata_name)
        create_dirs(FWUPD_DOM0_METADATA_DIR, FWUPD_DOM0_UNTRUSTED_DIR)
        with tempfile.TemporaryDirectory(dir=FWUPD_DOM0_UNTRUSTED_DIR) as tmpdir:
            cmd_copy_metadata_file = [
                "qvm-run",
                "--pass-io",
                "--no-gui",
                "--no-shell",
                "--",
                updatevm,
                "cat",
                "--",
                self.metadata_file_updatevm,
            ]
            # TODO: switch to ed25519 once firmware.xml.gz.jcat will have it
            cmd_copy_metadata_file_signature = [
                "qvm-run",
                "--pass-io",
                "--no-gui",
                "--no-shell",
                "--",
                updatevm,
                "cat",
                "--",
                self.metadata_file_updatevm + ".jcat",
            ]
            untrusted_metadata_file = os.path.join(tmpdir, metadata_name)

            with open(untrusted_metadata_file, "bx") as untrusted_file_1, open(
                untrusted_metadata_file + ".jcat", "bx"
            ) as untrusted_file_2, subprocess.Popen(
                cmd_copy_metadata_file, stdout=untrusted_file_1
            ) as p, subprocess.Popen(
                cmd_copy_metadata_file_signature, stdout=untrusted_file_2
            ) as q:
                p.wait()
                q.wait()
            if p.returncode != 0:
                raise Exception("qvm-run: Copying metadata file failed!!")
            if q.returncode != 0:
                raise Exception("qvm-run: Copying metadata jcat failed!!")

            self._jcat_verification(
                untrusted_metadata_file + ".jcat",
                os.path.dirname(untrusted_metadata_file),
                ignore_postquantum,
            )
            # verified, move into trusted dir
            shutil.move(untrusted_metadata_file, self.metadata_file)
            shutil.move(untrusted_metadata_file + ".jcat", self.metadata_file_jcat)

    def clean_cache(self):
        """Removes updates data"""
        print("Cleaning dom0 cache directories")
        if os.path.exists(FWUPD_DOM0_METADATA_DIR):
            shutil.rmtree(FWUPD_DOM0_METADATA_DIR)
        if os.path.exists(FWUPD_DOM0_UPDATES_DIR):
            shutil.rmtree(FWUPD_DOM0_UPDATES_DIR)
        if os.path.exists(HEADS_UPDATES_DIR):
            shutil.rmtree(HEADS_UPDATES_DIR)



#!/usr/bin/env python3
#
# Copyright 2021 Richard Hughes <richard@hughsie.com>
#
# SPDX-License-Identifier: LGPL-2.1-or-later
#
# pylint: disable=missing-module-docstring,missing-function-docstring,invalid-name,too-many-branches

import glob
import sys
from typing import List


def gtype_is_valid(gtype: str) -> bool:
    if not gtype:
        return False
    for token in ["Plugin", "Firmware", "Backend", "Image"]:
        if gtype.endswith(token):
            return False
    return True


def test_files() -> int:
    rc: int = 0

    gtypes: List[str] = []

    # find the possible GTypes
    for fn in glob.glob("libfwupdplugin/fu-*-device.c") + glob.glob("plugins/*/fu-*.c"):
        with open(fn) as f:
            for line in f.read().split("\n"):
                gtype: str = ""
                if line.startswith("G_DEFINE_TYPE("):
                    gtype = line[14:].split(",")[0]
                elif line.startswith("G_DEFINE_TYPE_WITH_PRIVATE("):
                    gtype = line[27:].split(",")[0]
                elif line.startswith("G_DEFINE_TYPE_WITH_CODE("):
                    gtype = line[24:].split(",")[0]
                if gtype_is_valid(gtype):
                    gtypes.append(gtype)

    # check each quirk file
    for fn in glob.glob("data/*.quirk") + glob.glob("plugins/*/*.quirk"):
        with open(fn) as f:
            for line in f.read().split("\n"):
                if line.startswith(" ") or line.endswith(" "):
                    print(f"{fn} has leading or trailing whitespace: {line}")
                    rc = 1
                    continue
                if not line or line.startswith("#"):
                    continue
                if line.startswith("["):
                    if not line.endswith("]"):
                        print(f"{fn} has invalid section header: {line}")
                        rc = 1
                        continue
                else:
                    sections = line.split(" = ")
                    if len(sections) != 2:
                        print(f"{fn} has invalid line: {line}")
                        rc = 1
                        continue
                    if sections[0] == "GType" and sections[1] not in gtypes:
                        print(f"{fn} has invalid GType: {line}")
                        rc = 1
                        continue
                    for section in sections:
                        if section.strip() != section:
                            print(f"{fn} has invalid spacing: {line}")
                            rc = 1
                            break
    return rc


if __name__ == "__main__":
    # all done!
    sys.exit(test_files())



#!/usr/bin/env python3
#
# The Qubes OS Project, http://www.qubes-os.org
#
# Copyright 2021 Norbert Kaminski <norbert.kaminski@3mdeb.com>
#
# SPDX-License-Identifier: LGPL-2.1-or-later
#

import json
import os
import shutil
import subprocess
import tempfile
import sys
import xml.etree.ElementTree as ET

from pathlib import Path
from packaging import version as pversion

FWUPD_QUBES_DIR = "/usr/share/qubes-fwupd"

# Check if script is run by tests and append sys path properly
if __name__ == "__main__":
    sys.path.append(os.path.join(FWUPD_QUBES_DIR, "src"))
else:
    sys.path.append("./src")

try:
    from qubes_fwupd_heads import FwupdHeads
    from qubes_fwupd_update import FwupdUpdate, run_in_tty
    from fwupd_receive_updates import FwupdReceiveUpdates
    from qubes_fwupd_common import EXIT_CODES, create_dirs
except ModuleNotFoundError:
    raise ModuleNotFoundError(
        "qubes-fwupd modules not found.  You may need to reinstall package."
    )

FWUPD_DOM0_DIR = "/var/cache/fwupd/qubes"
FWUPD_DOM0_METADATA_DIR = os.path.join(FWUPD_DOM0_DIR, "metadata")
FWUPD_DOM0_UPDATES_DIR = os.path.join(FWUPD_DOM0_DIR, "updates")
FWUPD_DOWNLOAD_PREFIX = "https://fwupd.org/downloads/"
METADATA_URL = "https://fwupd.org/downloads/firmware.xml.xz"
METADATA_URL_JCAT = "https://fwupd.org/downloads/firmware.xml.xz.jcat"

FWUPDMGR = "/bin/fwupdmgr"

BIOS_UPDATE_FLAG = os.path.join(FWUPD_DOM0_DIR, "bios_update")
LVFS_TESTING_DOM0_FLAG = os.path.join(FWUPD_DOM0_DIR, "lvfs_testing")

HELP = {
    "Usage": [
        {
            "Command": "qubes-fwupdmgr [OPTION…][FLAG..]",
            "Example": "qubes-fwupdmgr refresh --whonix --url=<url>\n",
        }
    ],
    "Options": [
        {
            "get-devices": "Get all devices that support firmware updates",
            "get-updates": "Get the list of updates for connected hardware",
            "get-remotes": "Get all enabled metadata remotes",
            "refresh": "Refresh metadata from remote server",
            "update": "Update chosen device to latest firmware version",
            "update-heads": "Updates heads firmware to the latest version (EXPERIMENTAL, not fully supported yet)",
            "downgrade": "Downgrade chosen device to chosen firmware version",
            "install": "Install firmware: DEVICE-UUID [VERSION] or --url with --sha",
            "clean": "Delete all cached update files\n",
        }
    ],
    "Flags": [
        {
            "--whonix": "Download firmware updates via Tor",
            "--device": "Specify device for heads update (default - x230)",
            "--url": "Address of the firmware or metadata remote server",
            "--sha": "SHA256 checksum of the firmware archive (required for install)",
            "--allow-older": "Allow installing an older firmware version",
            "--allow-reinstall": "Allow reinstalling the same firmware version",
            "--force": "Force the firmware update even if not required",
            "--ignore-postquantum": "Ignore pkcs7 failures for LVFS CA 2025PQ if at least one other CA signature is verified\n",
        }
    ],
    "Help": [{"-h --help": "Show help options\n"}],
}


class QubesFwupdmgr(FwupdHeads, FwupdUpdate, FwupdReceiveUpdates):
    def _download_metadata(
        self, whonix=False, metadata_url=None, ignore_postquantum=False
    ):
        """Initialize downloading metadata files.

        Keywords arguments:
        whonix -- Flag enforces downloading the metadata updates via Tor
        metadata_url -- Download metadata from the custom url
        """
        if not metadata_url:
            raise Exception("missing metadata URL")
        self.download_metadata(whonix=whonix, metadata_url=metadata_url)
        self.handle_metadata_update(
            self.updatevm,
            metadata_url=metadata_url,
            ignore_postquantum=ignore_postquantum,
        )
        if not os.path.exists(self.metadata_file):
            raise FileNotFoundError("Metadata file does not exist")

    def get_remotes(self):
        """Get metadata URLs for all enabled remotes"""

        if hasattr(self, "_remotes_cache"):
            return self._remotes_cache

        remotes_json = subprocess.check_output(
            [
                FWUPDMGR,
                "get-remotes",
                "--json",
            ]
        ).decode()
        remotes_list = json.loads(remotes_json)["Remotes"]
        remotes = {}
        for remote in remotes_list:
            name = remote["Id"]
            # fwupd 2.x emits "Enabled": true (bool), older versions used "true" (str)
            if remote.get("Enabled", True) not in (True, "true"):
                continue
            # skip local - for metadata refresh, we only care about those
            # actually needing refreshing
            if remote.get("Kind") != "download":
                continue
            assert "MetadataUri" in remote
            remotes[name] = remote["MetadataUri"]

        self._remotes_cache = remotes
        return self._remotes_cache

    def refresh_metadata(
        self,
        whonix=False,
        metadata_url=None,
        remote_name=None,
        ignore_postquantum=False,
    ):
        """Updates metadata with downloaded files.

        Keyword arguments:
        whonix -- Flag enforces downloading the metadata updates via Tor
        metadata_url -- Use custom metadata from the url
        remote_name -- Set refreshed metadata to this remote
        """
        if not metadata_url:
            if remote_name:
                metadata_url = self.get_remotes()[remote_name]
            else:
                raise Exception("missing metadata URL")
        metadata_name = os.path.basename(metadata_url)
        self.metadata_file = os.path.join(FWUPD_DOM0_METADATA_DIR, metadata_name)
        self.metadata_file_jcat = self.metadata_file + ".jcat"
        if not remote_name:
            if "testing" in metadata_url:
                remote_name = "lvfs-testing"
            else:
                remote_name = "lvfs"
        self._download_metadata(
            whonix=whonix,
            metadata_url=metadata_url,
            ignore_postquantum=ignore_postquantum,
        )
        cmd_refresh = [
            FWUPDMGR,
            "refresh",
            self.metadata_file,
            self.metadata_file_jcat,
            remote_name,
        ]
        p = subprocess.Popen(cmd_refresh, stdout=subprocess.PIPE)
        output = p.communicate()[0].decode()
        print(output)
        if p.returncode != 0:
            raise Exception("fwupd-qubes: Refresh failed")
        if "Successfully refreshed metadata manually" not in output:
            raise Exception(
                f"Manual metadata refresh failed: {output.strip() or '(no output)'}"
            )

    def refresh_metadata_all(self, whonix=False, ignore_postquantum=False):
        """Refresh metadata for all 'download' remotes

        Keyword arguments:
        whonix -- Flag enforces downloading the metadata updates via Tor
        """
        for name, url in self.get_remotes().items():
            try:
                self.refresh_metadata(
                    whonix=whonix,
                    remote_name=name,
                    metadata_url=url,
                    ignore_postquantum=ignore_postquantum,
                )
            except Exception as e:
                print(f"Failed to refresh remote '{name}': {e}")

    def _get_dom0_updates(self, allow_older=False, allow_reinstall=False):
        """Gathers information about available updates."""
        cmd_get_dom0_updates = [FWUPDMGR]
        if allow_older:
            cmd_get_dom0_updates.append("--allow-older")
        if allow_reinstall:
            cmd_get_dom0_updates.append("--allow-reinstall")
        cmd_get_dom0_updates += ["--json", "get-updates"]
        p = subprocess.Popen(cmd_get_dom0_updates, stdout=subprocess.PIPE)
        self.dom0_updates_info = p.communicate()[0].decode()
        if p.returncode != 0 and p.returncode != 2:
            raise Exception("fwupd-qubes: Getting available updates failed")

    def _parse_dom0_updates_info(self, updates_info):
        """Creates dictionary and list with information about updates.

        Keywords argument:
        updates_info - gathered update information
        """
        self.dom0_updates_info_dict = json.loads(updates_info)
        self.dom0_updates_list = [
            {
                "Name": device["Name"],
                "Version": device["Version"],
                "Releases": [
                    {
                        "Version": update["Version"],
                        "Url": update["Uri"],
                        "Checksum": update["Checksum"][-1],
                        "Description": update["Description"],
                    }
                    for update in device["Releases"]
                ],
            }
            for device in self.dom0_updates_info_dict["Devices"]
        ]

    def _download_firmware_updates(self, url, sha, whonix=False):
        """Initializes downloading firmware update archive.

        Keywords arguments:
        url -- url path to the firmware update archive
        sha -- SHA256 checksum of the firmware update archive
        whonix -- Flag enforces downloading the updates via Tor
        """
        self.cached = False
        self.download_firmware_updates(url, sha, whonix=whonix)
        if not self.cached:
            self.handle_fw_update(self.updatevm, sha, self.arch_name)
        if not os.path.exists(self.arch_path):
            raise FileNotFoundError("Firmware update files do not exist")

    def _user_input(self, updates_list, downgrade=False):
        """UI for update process.

        Keywords arguments:
        updates_dict - list of updates for specified device
        downgrade -- downgrade flag
        """
        decorator = "======================================================"
        if len(updates_list) == 0:
            print("No updates available.")
            return -EXIT_CODES["NOTHING_TO_DO"]
        if downgrade:
            print("Available downgrades:")
        else:
            print("Available updates:")
        self._updates_crawler(updates_list)

        while True:
            try:
                print("If you want to abandon process press 'N'.")
                choice = input("Otherwise choose a device number: ")
                if choice == "N" or choice == "n":
                    return -EXIT_CODES["NOTHING_TO_DO"]
                device_num = int(choice) - 1
                if 0 <= device_num < len(updates_list):
                    if not downgrade:
                        return device_num
                    break
                else:
                    raise ValueError()
            except ValueError:
                print("Invalid choice.")

        if downgrade:
            while True:
                try:
                    releases = updates_list[device_num]["Releases"]
                    for i, fw_dngd in enumerate(releases):
                        print(decorator)
                        print(
                            f"  {i+1}. Firmware downgrade version:"
                            f"\t {fw_dngd['Version']}"
                        )
                        description = fw_dngd["Description"].replace("<p>", "")
                        description = description.replace("<li>", "")
                        description = description.replace("<ul>", "")
                        description = description.replace("</ul>", "")
                        description = description.replace("</p>", "\n   ")
                        description = description.replace("</li>", "\n   ")
                        print(f"   Description:{description}")
                    print("If you want to abandon downgrade process press N.")
                    choice = input("Otherwise choose downgrade number: ")
                    if choice == "N" or choice == "n":
                        return -EXIT_CODES["NOTHING_TO_DO"]
                    downgrade_num = int(choice) - 1
                    if 0 <= downgrade_num < len(releases):
                        return device_num, downgrade_num
                    else:
                        raise ValueError()
                except ValueError:
                    print("Invalid choice.")

    def _parse_parameters(self, updates_list, choice):
        """Parses device name, url, version and SHA256 checksum of the file list.

        Keywords arguments:
        updates_list - list of updates for dom0
        choice -- number of device to be updated
        """
        self.name = updates_list[choice]["Name"]
        self.version = updates_list[choice]["Releases"][0]["Version"]
        for ver_check in updates_list[choice]["Releases"]:
            if pversion.parse(ver_check["Version"]) >= pversion.parse(self.version):
                self.version = ver_check["Version"]
                self.url = ver_check["Url"]
                self.sha = ver_check["Checksum"]

    def _read_dmi(self):
        """Reads BIOS information from DMI."""
        cmd_dmidecode_version = ["dmidecode", "-s", "bios-version"]
        p = subprocess.Popen(cmd_dmidecode_version, stdout=subprocess.PIPE)
        p.wait()
        self.dmi_version = p.communicate()[0].decode()
        cmd_dmidecode = ["dmidecode", "-t", "bios"]
        p = subprocess.Popen(cmd_dmidecode, stdout=subprocess.PIPE)
        p.wait()
        if p.returncode != 0:
            raise Exception("dmidecode: Reading DMI failed")
        return p.communicate()[0].decode()

    def _verify_dmi(self, arch_path, version, downgrade=False):
        """Verifies DMI tables for BIOS updates.

        Keywords arguments:
        arch_path -- absolute path of the update archive
        version -- version of the update
        downgrade -- downgrade flag
        """
        dmi_info = self._read_dmi()
        with tempfile.TemporaryDirectory() as tmpdir:
            cmd_extract = ["gcab", "-x", f"--directory={tmpdir}", "--", arch_path]
            p = subprocess.Popen(cmd_extract, stdout=subprocess.PIPE)
            p.communicate()
            if p.returncode != 0:
                raise Exception(f"gcab: Error while extracting {arch_path}.")
            path_metainfo = os.path.join(tmpdir, "firmware.metainfo.xml")
            tree = ET.parse(path_metainfo)
        root = tree.getroot()
        vendor = root.find("developer_name").text
        if vendor is None:
            raise ValueError("No vendor information in firmware metainfo.")
        if vendor not in dmi_info:
            raise ValueError("Wrong firmware provider.")
        if not downgrade and pversion.parse(version) <= pversion.parse(
            self.dmi_version
        ):
            raise ValueError(f"{version} < {self.dmi_version} Downgrade not allowed")

    def _get_dom0_devices(self):
        """Gathers information about devices connected in dom0."""
        cmd_get_dom0_devices = [FWUPDMGR, "--json", "get-devices"]
        p = subprocess.Popen(cmd_get_dom0_devices, stdout=subprocess.PIPE)
        self.dom0_devices_info = p.communicate()[0].decode()
        if p.returncode != 0:
            raise Exception("fwupd-qubes: Getting devices info failed")

    def update_firmware(
        self, whonix=False, allow_older=False, allow_reinstall=False, force=False
    ):
        """Updates firmware of the specified device.

        Keyword arguments:
        whonix -- Flag enforces downloading the metadata updates via Tor
        allow_older - allow installing an older firmware version
        allow_reinstall - allow reinstalling the same firmware version
        force - force installation even when not required
        """
        self._get_dom0_updates(allow_older=allow_older, allow_reinstall=allow_reinstall)
        self._parse_dom0_updates_info(self.dom0_updates_info)
        self._get_dom0_updates_string_based(
            allow_older=allow_older, allow_reinstall=allow_reinstall
        )
        updates_list = self.dom0_updates_list
        ret_input = self._user_input(updates_list)
        if ret_input == -EXIT_CODES["NOTHING_TO_DO"]:
            exit(EXIT_CODES["NOTHING_TO_DO"])
        choice = ret_input
        self._parse_parameters(updates_list, choice)
        self._download_firmware_updates(self.url, self.sha, whonix=whonix)
        if self.name == "System Firmware":
            Path(BIOS_UPDATE_FLAG).touch(mode=0o644, exist_ok=True)
            self._verify_dmi(self.arch_path, self.version)
        self._install_dom0_firmware(
            self.arch_path,
            allow_older=allow_older,
            allow_reinstall=allow_reinstall,
            force=force,
        )

    def _parse_downgrades(self, device_list):
        """Parses information about possible downgrades.

        Keywords argument:
        device_list -- list of connected devices
        """
        downgrades = []
        if "No detected devices" in device_list:
            return downgrades
        dom0_devices_info_dict = json.loads(device_list)
        for device in dom0_devices_info_dict["Devices"]:
            if "Releases" in device:
                try:
                    version = device["Version"]
                except KeyError:
                    continue
                downgrades.append(
                    {
                        "Name": device["Name"],
                        "Version": device["Version"],
                        "Releases": [
                            {
                                "Version": downgrade["Version"],
                                "Description": downgrade["Description"],
                                "Url": downgrade["Uri"],
                                "Checksum": downgrade["Checksum"][-1],
                            }
                            for downgrade in device["Releases"]
                            if pversion.parse(downgrade["Version"])
                            < pversion.parse(version)
                        ],
                    }
                )
        return downgrades

    def _install_dom0_firmware_downgrade(self, arch_path):
        """Installs firmware downgrade for specified device.

        Keywords arguments:
        arch_path - absolute path to firmware downgrade archive
        """
        cmd_install = [FWUPDMGR, "--allow-older", "install", arch_path]
        p = subprocess.Popen(cmd_install)
        p.wait()
        if p.returncode != 0:
            raise Exception("fwupd-qubes: Firmware downgrade failed")

    def _install_dom0_firmware(
        self, arch_path, allow_older=False, allow_reinstall=False, force=False
    ):
        """Installs firmware archive with optional install flags.

        Keywords arguments:
        arch_path - absolute path to firmware archive
        allow_older - pass --allow-older to fwupdmgr
        allow_reinstall - pass --allow-reinstall to fwupdmgr
        force - pass --force to fwupdmgr
        """
        cmd_install = [FWUPDMGR]
        if allow_older:
            cmd_install.append("--allow-older")
        if allow_reinstall:
            cmd_install.append("--allow-reinstall")
        if force:
            cmd_install.append("--force")
        cmd_install += ["install", arch_path]
        p = subprocess.Popen(cmd_install)
        p.wait()
        if p.returncode != 0:
            raise Exception(
                f"fwupd-qubes: Firmware install failed (fwupdmgr exit {p.returncode})"
            )

    def _find_device_release(
        self, device_id, version=None, allow_older=False, allow_reinstall=False
    ):
        """Look up firmware URL and SHA for a device by ID.

        Keywords arguments:
        device_id - DeviceId to identify the device
        version - exact firmware version to install, omit for best available
        allow_older - accept releases with a version older than current
        allow_reinstall - accept a release at the same version as current
        """
        self._get_dom0_devices()
        devices_dict = json.loads(self.dom0_devices_info)

        target_device = None
        for device in devices_dict.get("Devices", []):
            if device.get("DeviceId") == device_id:
                target_device = device
                break
        if target_device is None:
            raise Exception(f"Device '{device_id}' not found")

        current = target_device.get("Version", "0")
        releases = target_device.get("Releases", [])

        stderr = ""
        if not releases:
            releases, stderr = self._fwupd_get_releases(
                device_id, allow_older, allow_reinstall
            )
        if not releases:
            detail = f"\n  fwupdmgr: {stderr}" if stderr else ""
            raise Exception(
                f"No releases available for device '{device_id}'.{detail}\n"
                "  Try: sudo qubes-fwupdmgr refresh"
            )

        if version is not None:
            match = next((r for r in releases if r.get("Version") == version), None)
            if match is None:
                raise Exception(
                    f"Version '{version}' not found for device '{device_id}'"
                )
            uri = self._release_uri(match)
            if not uri:
                raise Exception(
                    f"Release '{version}' for device '{device_id}' has no download URL"
                )
            return uri, match["Checksum"][-1]

        current = target_device.get("Version", "0")
        candidates = [
            r
            for r in releases
            if self._release_eligible(
                r.get("Version"), current, allow_older, allow_reinstall
            )
        ]
        if not candidates:
            raise Exception(f"No eligible release found for device '{device_id}'")
        best = max(candidates, key=lambda r: pversion.Version(r["Version"]))
        uri = self._release_uri(best)
        if not uri:
            raise Exception(
                f"Best release for device '{device_id}' has no download URL"
            )
        return uri, best["Checksum"][-1]

    def install_firmware(
        self,
        device_id=None,
        version=None,
        url=None,
        sha=None,
        allow_older=False,
        allow_reinstall=False,
        force=False,
        whonix=False,
    ):
        """Downloads and installs firmware for a device or from a direct URL.

        Keyword arguments:
        device_id - DeviceId or GUID to identify the target device
        version - specific firmware version to install (used with device_id)
        url - direct LVFS URL of the firmware cabinet (alternative to device_id)
        sha - SHA256 checksum matching url
        allow_older - allow installing an older firmware version
        allow_reinstall - allow reinstalling the same firmware version
        force - force installation even when not required
        whonix - route download through sys-whonix (Tor)
        """
        if device_id is not None:
            url, sha = self._find_device_release(
                device_id,
                version=version,
                allow_older=allow_older,
                allow_reinstall=allow_reinstall,
            )
        elif not (url and sha):
            raise Exception("install requires a device UUID or --url and --sha")
        self._download_firmware_updates(url, sha, whonix=whonix)
        self._install_dom0_firmware(
            self.arch_path,
            allow_older=allow_older,
            allow_reinstall=allow_reinstall,
            force=force,
        )

    def downgrade_firmware(self, whonix=False):
        """Downgrades firmware of the specified device.

        Keyword arguments:
        whonix -- Flag enforces downloading the metadata updates via Tor
        """
        self._get_dom0_devices()
        dom0_downgrades = self._parse_downgrades(self.dom0_devices_info)
        ret_input = self._user_input(dom0_downgrades, downgrade=True)
        if ret_input == -EXIT_CODES["NOTHING_TO_DO"]:
            exit(EXIT_CODES["NOTHING_TO_DO"])
        device_choice, downgrade_choice = ret_input
        downgrade = dom0_downgrades[device_choice]
        releases = downgrade["Releases"]
        downgrade_url = releases[downgrade_choice]["Url"]
        downgrade_sha = releases[downgrade_choice]["Checksum"]
        self._download_firmware_updates(downgrade_url, downgrade_sha, whonix=whonix)
        if downgrade["Name"] == "System Firmware":
            Path(BIOS_UPDATE_FLAG).touch(mode=0o644, exist_ok=True)
            self._verify_dmi(
                self.arch_path,
                downgrade["Version"],
                downgrade=True,
            )
        self._install_dom0_firmware_downgrade(self.arch_path)

    def _output_crawler(self, updev_dict, level, help_f=False):
        """Prints device and updates information as a tree.

        Keywords arguments:
        updev_dict -- update/device information dictionary
        level -- level of the tree
        """

        def _tabs(key_word):
            return key_word + "\t" * (4 - int(len(key_word) / 8))

        decorator = "==================================="
        print(2 * decorator)
        for updev_key in updev_dict:
            style = "\t" * level
            output = style + _tabs(updev_key + ":")
            if len(updev_key) > 20:
                continue
            if updev_key == "Icons":
                continue
            if updev_key == "Releases":
                continue
            if updev_key == "Name":
                print(style + updev_dict["Name"])
                print(2 * decorator)
                continue
            if isinstance(updev_dict[updev_key], str):
                print(output + updev_dict[updev_key])
            elif isinstance(updev_dict[updev_key], int):
                print(output + str(updev_dict[updev_key]))
            elif isinstance(updev_dict[updev_key][0], str):
                for i, data in enumerate(updev_dict[updev_key]):
                    if i == 0:
                        print(output + "\u00b7" + data)
                        continue
                    print(style + _tabs(" ") + "\u00b7" + data)
            elif isinstance(updev_dict[updev_key][0], dict):
                if level == 0 and help_f is True:
                    print(output)
                else:
                    if level == 0:
                        print(f"Dom0 {output}")

                for nested_dict in updev_dict[updev_key]:
                    self._output_crawler(nested_dict, level + 1)

    def _updates_crawler(self, updates_list, prefix=0):
        """Prints updates information for dom0

        Keywords arguments:
        updates_list -- list of devices updates
        prefix -- device number prefix
        """
        available_updates = False
        decorator = "======================================================"
        print(decorator)
        print("Dom0 updates:")
        print(decorator)
        if len(updates_list) == 0:
            print("No updates available.")
            return EXIT_CODES["NOTHING_TO_DO"]
        else:
            for i, device in enumerate(updates_list):
                if len(device["Releases"]) == 0:
                    continue
                if not available_updates:
                    print("Available updates:")
                    print(decorator)
                print("^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^")
                print(f"{i+1+prefix}. Device: {device['Name']}")
                print(f"   Current firmware version:\t {device['Version']}")
                for update in device["Releases"]:
                    print(decorator)
                    print("   Firmware update " f"version:\t {update['Version']}")
                    print(f"   URL:\t {update['Url']}")
                    print(f"   SHA256 checksum:\t {update['Checksum']}")
                    description = update["Description"].replace("<p>", "")
                    description = description.replace("<li>", "")
                    description = description.replace("<ul>", "")
                    description = description.replace("</ul>", "")
                    description = description.replace("</p>", "\n\t")
                    description = description.replace("</li>", "\n\t")
                    print(f"   Description: {description}")
                print(decorator)
                available_updates = True
            if not available_updates:
                print("No updates available.")
                return EXIT_CODES["NOTHING_TO_DO"]

    def get_devices_qubes(self):
        """Gathers and prints devices information."""
        self._get_dom0_devices()
        dom0_devices_info_dict = json.loads(self.dom0_devices_info)
        self._output_crawler(dom0_devices_info_dict, 0)

    def get_remotes_qubes(self):
        """Prints enabled download remotes and their metadata URIs."""
        remotes = self.get_remotes()
        if not remotes:
            print("No enabled download remotes found.")
            return
        decorator = "======================================================"
        print(decorator)
        print("Enabled remotes:")
        print(decorator)
        for name, uri in remotes.items():
            print(f"  {name}: {uri}")

    def get_updates_qubes(self, allow_older=False, allow_reinstall=False):
        """Gathers and prints updates information."""
        self._get_dom0_updates(allow_older=allow_older, allow_reinstall=allow_reinstall)
        self._parse_dom0_updates_info(self.dom0_updates_info)
        self._get_dom0_updates_string_based(
            allow_older=allow_older, allow_reinstall=allow_reinstall
        )
        self._updates_crawler(self.dom0_updates_list)

    def _get_dom0_updates_string_based(self, allow_older=False, allow_reinstall=False):
        devices_with_releases = sum(1 for d in self.dom0_updates_list if d["Releases"])
        if devices_with_releases == 0:
            self.dom0_updates_list = self._string_based_updates(
                allow_older=allow_older,
                allow_reinstall=allow_reinstall,
            )

    def help(self):
        """Prints help information"""
        self._output_crawler(HELP, 0, help_f=True)

    def trusted_cleanup(self):
        """Deletes trusted directory."""
        trusted_path = os.path.join(FWUPD_DOM0_UPDATES_DIR, "trusted.cab")
        if os.path.exists(trusted_path):
            os.remove(trusted_path)
            shutil.rmtree(trusted_path.replace(".cab", ""))

    def refresh_metadata_after_bios_update(self):
        """Refreshes metadata after bios update"""
        if os.path.exists(BIOS_UPDATE_FLAG):
            print("BIOS was updated. Refreshing metadata...")
            if "--whonix" in sys.argv:
                self.refresh_metadata_all(whonix=True)
            else:
                self.refresh_metadata_all()
            os.remove(BIOS_UPDATE_FLAG)

    def heads_update(
        self, device=None, whonix=False, metadata_url=None, ignore_postquantum=False
    ):
        """
        Updates heads firmware

        Keyword arguments:
        device -- Model of the updated device
        whonix -- Flag enforces downloading the metadata updates via Tor
        metadata_url -- Use custom metadata from the url
        """
        if not metadata_url:
            metadata_url = METADATA_URL
        metadata_name = os.path.basename(metadata_url)
        self.metadata_file = os.path.join(FWUPD_DOM0_METADATA_DIR, metadata_name)
        self._get_hwids()
        if device is None:
            device = self._get_hwid_device()
        self._download_metadata(
            whonix=whonix,
            metadata_url=metadata_url,
            ignore_postquantum=ignore_postquantum,
        )
        self._parse_metadata(self.metadata_file)
        if self._gather_firmware_version() == EXIT_CODES["NOTHING_TO_DO"]:
            return EXIT_CODES["NOTHING_TO_DO"]
        if self._parse_heads_updates(device) == EXIT_CODES["NOTHING_TO_DO"]:
            return EXIT_CODES["NOTHING_TO_DO"]
        self._download_firmware_updates(self.heads_update_url, self.heads_update_sha)
        return_code = self._copy_heads_firmware(self.arch_path)
        if return_code == EXIT_CODES["NOTHING_TO_DO"]:
            exit(EXIT_CODES["NOTHING_TO_DO"])
        elif return_code == EXIT_CODES["SUCCESS"]:
            print()
            while True:
                try:
                    print("An update requires a reboot to complete.")
                    choice = input("Do you want to restart now? (Y|N)\n")
                    if choice == "N" or choice == "n":
                        return EXIT_CODES["SUCCESS"]
                    elif choice == "Y" or choice == "y":
                        print("Rebooting...")
                        os.system("reboot")
                    else:
                        raise ValueError()
                except ValueError:
                    print("Invalid choice.")
        else:
            raise Exception("Copying heads update failed!!")

    def validate_dom0_dirs(self):
        """Validates and creates directories"""
        if not os.path.exists(FWUPD_DOM0_DIR):
            create_dirs(FWUPD_DOM0_DIR)
        if os.path.exists(FWUPD_DOM0_METADATA_DIR):
            shutil.rmtree(FWUPD_DOM0_METADATA_DIR)
            create_dirs(FWUPD_DOM0_METADATA_DIR)
        else:
            create_dirs(FWUPD_DOM0_METADATA_DIR)
        if not os.path.exists(FWUPD_DOM0_UPDATES_DIR):
            create_dirs(FWUPD_DOM0_UPDATES_DIR)

    def _release_uri(self, release):
        """Return the download URL from a release dict.

        Older fwupd emits 'Uri' (string), newer fwupd emits 'Locations' (list).
        """
        uri = release.get("Uri")
        if uri:
            return uri
        locations = release.get("Locations")
        if locations:
            return locations[0]
        return None

    def _string_based_updates(self, allow_older=False, allow_reinstall=False):
        """Fallback update scan using version-string comparison."""
        self._get_dom0_devices()
        devices = json.loads(self.dom0_devices_info).get("Devices", [])

        updates_list = []
        for device in devices:
            if "updatable" not in device.get("Flags", []):
                continue
            dev_id = device.get("DeviceId")
            if not dev_id:
                continue
            dev_ver = device.get("Version", "0")

            releases, _ = self._fwupd_get_releases(dev_id, allow_older, allow_reinstall)
            candidates = [
                r
                for r in releases
                if self._release_eligible(
                    r.get("Version"), dev_ver, allow_older, allow_reinstall
                )
            ]
            if not candidates:
                continue

            updates_list.append(
                {
                    "Name": device.get("Name", "Unknown"),
                    "Version": dev_ver,
                    "Releases": [
                        {
                            "Version": r["Version"],
                            "Url": self._release_uri(r),
                            "Checksum": r["Checksum"][-1],
                            "Description": r.get("Description", ""),
                        }
                        for r in candidates
                        if self._release_uri(r) and r.get("Checksum")
                    ],
                }
            )
        return updates_list

    def _release_eligible(self, rel_ver_str, cur_ver_str, allow_older, allow_reinstall):
        """Return True if a release is eligible to install over current using
        packaging.version human-readable Version strings.
        """
        if not rel_ver_str:
            return False
        try:
            rv = pversion.parse(rel_ver_str)
            cv = pversion.parse(cur_ver_str or "0")
        except Exception:
            return False
        if rv > cv:
            return True
        if rv == cv and allow_reinstall:
            return True
        if rv < cv and allow_older:
            return True
        return False

    def _fwupd_get_releases(self, device_id, allow_older=False, allow_reinstall=False):
        """Run `fwupdmgr --json get-releases <id>` and return its Releases list.

        Returns ([], stderr_str) on failure so the caller can build a useful
        error message. stderr_str is "" when nothing went wrong.
        """
        cmd = [FWUPDMGR, "--json"]
        if allow_older:
            cmd.append("--allow-older")
        if allow_reinstall:
            cmd.append("--allow-reinstall")
        cmd += ["get-releases", "--", device_id]
        try:
            result = subprocess.run(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        except Exception as e:
            return [], str(e)
        try:
            releases = json.loads(result.stdout.decode(errors="replace")).get(
                "Releases", []
            )
        except (json.JSONDecodeError, UnicodeDecodeError):
            releases = []
        stderr = result.stderr.decode(errors="replace").strip()
        return releases, stderr


def main():
    if os.geteuid() != 0:
        print("You need to have root privileges to run this script.\n")
        exit(EXIT_CODES["ERROR"])

    q = QubesFwupdmgr()

    if len(sys.argv) < 2:
        q.help()
        exit(1)

    metadata_url = None
    device_override = None
    firmware_sha = None
    whonix = False
    allow_older = False
    allow_reinstall = False
    force = False
    ignore_postquantum = False

    for arg in sys.argv:
        if "--url=" in arg:
            metadata_url = arg.replace("--url=", "")
            if not metadata_url.startswith(FWUPD_DOWNLOAD_PREFIX):
                print(
                    "Metadata must be stored in the Linux"
                    " Vendor Firmware Service (https://fwupd.org/)"
                )
                print("Exiting...")
                exit(1)
        if "--sha=" in arg:
            firmware_sha = arg.replace("--sha=", "")
        if "--device=" in arg:
            device_override = arg.replace("--device=", "")
        if "--whonix" == arg:
            whonix = True
        if "--allow-older" == arg:
            allow_older = True
        if "--allow-reinstall" == arg:
            allow_reinstall = True
        if "--force" == arg:
            force = True
        if "--ignore-postquantum" == arg:
            ignore_postquantum = True

    q.validate_dom0_dirs()
    q.trusted_cleanup()
    q.refresh_metadata_after_bios_update()

    if not os.path.exists(FWUPD_DOM0_DIR):
        if metadata_url:
            q.refresh_metadata(
                whonix=whonix,
                metadata_url=metadata_url,
                ignore_postquantum=ignore_postquantum,
            )
        else:
            q.refresh_metadata_all(whonix=whonix)

    if sys.argv[1] == "get-updates":
        q.get_updates_qubes(allow_older=allow_older, allow_reinstall=allow_reinstall)
    elif sys.argv[1] == "get-devices":
        q.get_devices_qubes()
    elif sys.argv[1] == "get-remotes":
        q.get_remotes_qubes()
    elif sys.argv[1] == "update":
        q.update_firmware(
            whonix=whonix,
            allow_older=allow_older,
            allow_reinstall=allow_reinstall,
            force=force,
        )
    elif sys.argv[1] == "downgrade":
        q.downgrade_firmware(whonix=whonix)
    elif sys.argv[1] == "install":
        install_pos = [a for a in sys.argv[2:] if not a.startswith("--")]
        install_device_id = install_pos[0] if install_pos else None
        install_version = install_pos[1] if len(install_pos) > 1 else None
        if install_device_id and metadata_url:
            print("install: use either DEVICE-UUID or --url, not both")
            exit(EXIT_CODES["ERROR"])
        if not install_device_id and not (metadata_url and firmware_sha):
            print(
                "install requires DEVICE-UUID [VERSION] or --url=<URL> --sha=<SHA256>"
            )
            exit(EXIT_CODES["ERROR"])
        q.install_firmware(
            device_id=install_device_id,
            version=install_version,
            url=metadata_url,
            sha=firmware_sha,
            allow_older=allow_older,
            allow_reinstall=allow_reinstall,
            force=force,
            whonix=whonix,
        )
    elif sys.argv[1] == "clean":
        q.clean_cache()
    elif sys.argv[1] == "refresh":
        if metadata_url:
            q.refresh_metadata(
                whonix=whonix,
                metadata_url=metadata_url,
                ignore_postquantum=ignore_postquantum,
            )
        else:
            q.refresh_metadata_all(whonix=whonix, ignore_postquantum=ignore_postquantum)
    elif sys.argv[1] == "update-heads":
        q.heads_update(
            device=device_override,
            metadata_url=metadata_url,
            whonix=whonix,
            ignore_postquantum=ignore_postquantum,
        )
    else:
        q.help()
        exit(1)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
#
# Copyright 2026 Richard Hughes <richard@hughsie.com>
#
# SPDX-License-Identifier: LGPL-2.1-or-later
#
# pylint: disable=invalid-name,missing-docstring

import sys
import io
import zipfile
import tarfile

try:
    fn_old = sys.argv[1]
except IndexError:
    print("USAGE: ./convert-to-zip.py INPUT.dat [OUTPUT.zip]")
    sys.exit(1)
try:
    fn_new = sys.argv[2]
except IndexError:
    fn_new = fn_old.replace(".dat", ".zip")
with zipfile.ZipFile(fn_new, "w", zipfile.ZIP_DEFLATED) as zipf:
    with tarfile.open(fn_old, "r:bz2") as tar_outer:
        safe_isp: bool = False
        for i in tar_outer:
            if i.name == "update/cx3uvc.img":
                safe_isp = True
        if not safe_isp:
            print("This script was designed for SafeISP images only")
            sys.exit(1)
        for i in tar_outer:
            print(f"extracting {i.name}")
            tarfile_outer = tar_outer.extractfile(i)
            if not tarfile_outer:
                print(f"ignoring {i.name}")
                continue
            fileobj = io.BytesIO(tarfile_outer.read())
            with tarfile.open(fileobj=fileobj, mode="r:bz2") as tar_inner:
                for j in tar_inner:
                    tarfile_inner = tar_inner.extractfile(j)
                    if not tarfile_inner:
                        print(f"ignoring {j.name}")
                        continue
                    print(f"adding {j.name}...")
                    zipf.writestr(j.name, data=tarfile_inner.read())

print(f"saved: {fn_new}")



#!/usr/bin/env python3
#
# Copyright 2021 Richard Hughes <richard@hughsie.com>
# Copyright 2021 Mario Limonciello <superm1@gmail.com>
#
# SPDX-License-Identifier: LGPL-2.1-or-later

import glob
import os
import sys
import fnmatch
from typing import List


def test_files() -> int:
    rc: int = 0
    build_dirs = [os.path.dirname(cf) for cf in glob.glob("**/config.h")]

    for fn in (
        glob.glob("libfwupd/*.[c|h|py|rs]")
        + glob.glob("libfwupdplugin/.[c|h|py|rs]")
        + glob.glob("plugins/*/.[c|h|py|rs]")
        + glob.glob("src/.[c|h|py|rs]")
    ):
        if fn.endswith("check-license.py"):
            continue
        lic: str = ""
        cprts: List[str] = []
        lines: List[str] = []
        with open(fn) as f:
            for line in f.read().split("\n"):
                lines.append(line)
        if len(lines) < 2:
            continue
        for line in lines:
            if line.find("SPDX-License-Identifier:") != -1:
                lic = line.split(":")[1]
            if line.find("Copyright") != -1:
                cprts.append(line.strip())
        if not lic:
            print(f"{fn} does not specify a license")
            rc = 1
            continue
        if not cprts:
            print(f"{fn} does not specify any copyright")
            rc = 1
            continue
        if "LGPL-2.1-or-later" not in lic:
            print(f"{fn} does not contain LGPL-2.1-or-later ({lic})")
            rc = 1
            continue
        for cprt in cprts:
            for word in ["(C)", "(c)", "©", "  "]:
                if cprt.find(word) != -1:
                    print(f"{fn} should not contain {word} in the string {cprt}")
                    rc = 1
    return rc


if __name__ == "__main__":
    # all done!
    sys.exit(test_files())



#!/usr/bin/env python3
#
# Copyright 2023 Richard Hughes <richard@hughsie.com>
#
# SPDX-License-Identifier: LGPL-2.1-or-later
#
# pylint: disable=invalid-name,missing-module-docstring,missing-function-docstring
# pylint: disable=too-many-lines,too-many-return-statements,missing-class-docstring
# pylint: disable=too-many-arguments,too-many-positional-arguments,too-many-statements
# pylint: disable=too-few-public-methods,too-many-branches,protected-access

import glob
import sys
import os
import argparse
from typing import List, Optional
from ctokenizer import Tokenizer, Node, NodeHint, Token, TokenHint


# convert a CamelCase name into snake_case
def _camel_to_snake(name: str) -> str:
    name_snake: str = ""
    for char in name:
        if char.islower() or char.isnumeric():
            name_snake += char
            continue
        if char == "_":
            name_snake += char
            continue
        if name_snake:  # and not char.isnumeric():
            name_snake += "_"
        name_snake += char.lower()
    return name_snake


def _value_relaxed(data: str) -> str:
    if data in ["0x0", "0x00", "0x0000"]:
        return "0"
    if data in ["G_SOURCE_REMOVE"]:
        return "FALSE"
    if data in ["G_SOURCE_CONTINUE"]:
        return "TRUE"
    return data


class SourceFailure:
    def __init__(
        self, fn=None, linecnt=None, message=None, nocheck=None, expected=False
    ) -> None:
        self.fn: Optional[str] = fn
        self.linecnt: Optional[int] = linecnt
        self.message: Optional[str] = message
        self.nocheck: Optional[str] = nocheck
        self.expected: bool = expected


class Checker:

    def __init__(self, verbose: bool = False) -> None:
        self.verbose: bool = verbose
        self.failures: List[SourceFailure] = []
        self._current_fn: Optional[str] = None
        self._current_nocheck: Optional[str] = None
        self._gtype_parents: dict[str, str] = {}
        self._klass_funcs: List[str] = []
        self._expected_failure_prefixes: List[str] = []

    def add_expected_failure(self, message_prefix: str) -> None:
        self._expected_failure_prefixes.append(message_prefix)

    def _should_process_node(self, node: Node) -> bool:

        if self._current_nocheck:
            if node.tokens_pre.count_fuzzy([f"~/* {self._current_nocheck} */"]) > 0:
                return False
            if node.tokens.count_fuzzy([f"~/* {self._current_nocheck} */"]) > 0:
                return False
        return True

    def add_failure(self, message=None, linecnt: Optional[int] = None) -> None:

        # we were expecting this
        expected: bool = False
        for message_prefix in self._expected_failure_prefixes:
            if message.startswith(message_prefix):
                expected = True
                break

        # add
        self.failures.append(
            SourceFailure(
                expected=expected,
                fn=self._current_fn,
                linecnt=linecnt,
                message=message,
                nocheck=self._current_nocheck,
            )
        )

    def has_failure(self, message_prefix: str) -> bool:
        for failure in self.failures:
            if failure.message and failure.message.startswith(message_prefix):
                return True
        return False

    def _test_function_names_prefix_private(self, func_name: str, token: Token) -> None:

        if func_name.startswith("__"):
            return
        valid_prefixes = ["_fwupd_", "_fu_", "_g_", "_xb_"]
        for prefix in valid_prefixes:
            if func_name.startswith(prefix):
                return
        self.add_failure(
            f"invalid function name {func_name} should have {'|'.join(valid_prefixes)} prefix",
            linecnt=token.linecnt,
        )

    def _test_function_names_prefix_valid(self, func_name: str, token: Token) -> None:
        # sanity check
        if not self._current_fn:
            return

        # ignore headers
        if self._current_fn.endswith(".h"):
            return

        # ignore all self tests FIXME: 335 failures
        if self._current_fn.endswith("-test.c"):
            return

        # namespacing is strange here
        if self._current_fn.endswith("fwupd-enums.c"):
            return

        # doh
        if func_name in ["main", "fu_plugin_init_vfuncs", "fu_plugin_get_data"]:
            return

        # this is stuff that should move to GLib
        if func_name.startswith("_g_"):
            return

        # remove some suffixes we do not care about
        prefix = os.path.basename(self._current_fn).split(".")[0].replace("-", "_")
        for suffix in [
            "_common",
            "_darwin",
            "_freebsd",
            "_helper",
            "_impl",
            "_linux",
            "_sync",
            "_windows",
        ]:
            if prefix.endswith(suffix):
                prefix = prefix[: -len(suffix)]

        # allowed truncations
        valid_prefixes = []
        valid_prefixes.append(prefix)
        for key, value in {
            "fu_crc": "fu_misr",  # split out to fu-misr.[c|h]?
            "fu_darwin_efivars": "fu_efivars",
            "fu_dbus_daemon": "fu_daemon",
            "fu_dbxtool": "fu_util",
            "fu_freebsd_efivars": "fu_efivars",
            "fu_linux_efivars": "fu_efivars",
            "fu_logitech_hidpp_hidpp": "fu_logitech_hidpp",
            "fu_logitech_hidpp_hidpp_msg": "fu_logitech_hidpp",
            "fu_self_test": "fu_test",
            "fu_string": "fu_str,fu_utf",
            "fu_tool": "fu_util",
            "fu_windows_efivars": "fu_efivars",
        }.items():
            if prefix == key:
                valid_prefixes.extend(value.split(","))
        for prefix in valid_prefixes:
            if func_name.startswith(prefix):
                return

        self.add_failure(
            f"invalid function name {func_name} should have {'|'.join(valid_prefixes)} prefix",
            linecnt=token.linecnt,
        )

    def _discover_klass_functions(self, node: Node) -> None:
        """discover all the device_class-> functions"""

        if node.depth != 0:
            return
        idx = node.tokens_pre.find_fuzzy(["~fu_*_class_init@FUNCTION"])
        if idx == -1:
            return

        idx = 0
        while True:
            idx = node.tokens.find_fuzzy(["-", ">", "~*", "="], offset=idx + 1)
            if idx == -1:
                break
            self._klass_funcs.append(node.tokens[idx + 4].data)

    def _test_variable_case(self, node: Node) -> None:
        """disallow non-lowercase variables"""

        if node.hint == NodeHint.ENUM:
            return

        idx: int = 0
        while True:
            idx = node.tokens.find_fuzzy(["="], offset=idx + 1)
            if idx == -1:
                return
            token = node.tokens[idx - 1]
            if token.data.find(".") != -1:
                continue
            if token.data.lower() != token.data:
                self.add_failure(
                    f"mixed case variable {token.data}",
                    linecnt=token.linecnt,
                )

    def _test_variable_buffer_index(self, node: Node) -> None:
        """disallow buf[] and prefer *buf"""

        idx: int = 0
        while True:
            idx = node.tokens_pre.find_fuzzy(
                ["guint8", "~*", "[", "]", ","], offset=idx + 1
            )
            if idx == -1:
                return
            token = node.tokens_pre[idx + 1]
            self.add_failure(
                f"do not use param guint8 {token.data}[], instead use 'guint8 *{token.data}'",
                linecnt=token.linecnt,
            )

    def _test_struct_member_case(self, node: Node) -> None:
        """disallow non-lowercase struct members"""

        if node.hint not in [NodeHint.STRUCT, NodeHint.STRUCT_TYPEDEF]:
            return
        idx: int = 0
        while True:
            idx = node.tokens.find_fuzzy([";"], offset=idx + 1)
            if idx == -1:
                return
            token = node.tokens[idx - 1]
            if token.data.lower() != token.data:
                self.add_failure(
                    f"mixed case struct member {token.data}",
                    linecnt=token.linecnt,
                )

    def _test_param_self_native_device(self, node: Node) -> None:
        """use @self for native internal functions, not the basetype"""

        if self._current_fn and os.path.basename(self._current_fn) in [
            "fu-device.c",
            "fu-device-locker.c",
            "fu-device-progress.c",
        ]:
            return
        if self._current_fn and os.path.basename(self._current_fn).endswith(".h"):
            return
        if node.depth != 0:
            return
        idx = node.tokens_pre.find_fuzzy(["@FUNCTION", "(", "FuDevice", "~*"])
        if idx == -1:
            return
        token = node.tokens_pre[idx]
        if token.data.endswith("_cb"):
            return
        if token.data in self._klass_funcs:
            return
        token_param = node.tokens_pre[idx + 4]
        if token_param.data == "self":
            return
        if token.data.endswith("_new") and token_param.data == "proxy":
            return
        self.add_failure(
            f"native device functions should use self as the first parameter not {token_param.data}",
            linecnt=token.linecnt,
        )

    def _test_param_self_native_firmware(self, node: Node) -> None:
        """use @self for native internal functions, not the basetype"""

        if self._current_fn and os.path.basename(self._current_fn) in [
            "fu-firmware-progress.c",
            "fu-self-test.c",
        ]:
            return
        if self._current_fn and os.path.basename(self._current_fn).endswith(".h"):
            return
        if self._current_fn and os.path.basename(self._current_fn).endswith(
            "-common.c"
        ):
            return
        if node.depth != 0:
            return
        idx = node.tokens_pre.find_fuzzy(
            ["@FUNCTION", "(", "FuFirmware", "~*", "firmware"]
        )
        if idx == -1:
            return
        token = node.tokens_pre[idx]
        if token.data in self._klass_funcs:
            return
        self.add_failure(
            "native firmware functions should use self as the first parameter not firmware",
            linecnt=token.linecnt,
        )

    def _test_param_self_device(self, node: Node) -> None:
        """only use @self for device GTypes, not the basetype"""

        if self._current_fn and os.path.basename(self._current_fn) in [
            "fu-device.c",
            "fu-device.h",
            "fu-device-poll-locker.h",
            "fu-device-private.h",
        ]:
            return
        if node.depth != 0:
            return
        idx = node.tokens_pre.find_fuzzy(["@FUNCTION", "(", "FuDevice", "*", "self"])
        if idx == -1:
            return
        token = node.tokens_pre[idx + 4]
        self.add_failure(
            f"invalid parameter name {token.data} should be called 'device'",
            linecnt=token.linecnt,
        )

    def _test_param_self_firmware(self, node: Node) -> None:
        """only use @self for firmware GTypes, not the basetype"""

        if self._current_fn and os.path.basename(self._current_fn) in [
            "fu-firmware.c",
            "fu-firmware-private.h",
        ]:
            return
        if node.depth != 0:
            return
        idx = node.tokens_pre.find_fuzzy(["@FUNCTION", "(", "FuFirmware", "*", "self"])
        if idx == -1:
            return
        token = node.tokens_pre[idx + 4]
        self.add_failure(
            f"invalid parameter name {token.data} should be called 'firmware'",
            linecnt=token.linecnt,
        )

    def _test_function_names_ensure(self, node: Node) -> None:
        """setting internal state should be 'ensuring' the data, not setting it"""

        if node.depth != 0:
            return

        idx = node.tokens_pre.find_fuzzy(
            [
                "gboolean",
                "~fu_*_set_*@FUNCTION",
                "(",
                "~*",
                "*",
                "self",
                ",",
                "GError",
                "*",
                "*",
                "error",
                ")",
            ]
        )
        if idx != -1:
            token = node.tokens_pre[idx]
            self.add_failure(
                "function should be called ensure, not set",
                linecnt=token.linecnt,
            )

    def _test_function_name_suffix(self, node: Node) -> None:
        """verb_noun, not noun_verb"""

        if node.depth != 0:
            return

        # ACTION_set
        idx = node.tokens_pre.find_fuzzy(["~fu_*_set@FUNCTION"])
        if idx != -1:
            token = node.tokens_pre[idx]
            self.add_failure(
                "function should be called set_ACTION, not ACTION_set",
                linecnt=token.linecnt,
            )

        # ACTION_get
        idx = node.tokens_pre.find_fuzzy(["~fu_*_get@FUNCTION"])
        if idx != -1:
            token = node.tokens_pre[idx]
            self.add_failure(
                "function should be called get_ACTION, not ACTION_get",
                linecnt=token.linecnt,
            )

        # THING_add
        idx = node.tokens_pre.find_fuzzy(["~fu_*_add@FUNCTION"])
        if idx != -1:
            token = node.tokens_pre[idx]
            if (
                token.data.find("_list_") == -1
                and token.data.find("_array_") == -1
                and token.data.find("_cache_") == -1
                and token.data not in ["fu_size_checked_add"]
            ):
                self.add_failure(
                    "function should be called add_THING, not THING_add",
                    linecnt=token.linecnt,
                )

        # THING_remove
        idx = node.tokens_pre.find_fuzzy(["~fu_*_remove@FUNCTION"])
        if idx != -1:
            token = node.tokens_pre[idx]
            if token.data.find("_list_") == -1 and token.data.find("_cache_") == -1:
                self.add_failure(
                    "function should be called remove_THING, not THING_remove",
                    linecnt=token.linecnt,
                )

    def _test_function_names_prefix(self, node: Node) -> None:

        if node.depth != 0:
            return

        idx: int = 0
        func_name: str = ""
        while True:
            idx = node.tokens_pre.find_fuzzy(["@FUNCTION", "("], offset=idx + 1)
            if idx == -1:
                return

            # sanity check
            token = node.tokens_pre[idx]
            func_name = token.data
            if len(func_name) < 10:
                continue
            if func_name in ["__attribute__"]:
                continue
            if self.verbose:
                print("func_name", func_name)
            if func_name.upper() != func_name and func_name.lower() != func_name:
                self.add_failure(
                    "mixed case function name",
                    linecnt=token.linecnt,
                )
            if func_name.startswith("_"):
                self._test_function_names_prefix_private(func_name, token)
            else:
                self._test_function_names_prefix_valid(func_name, token)

    def _test_missing_literal(self, node: Node) -> None:
        """test for missing literals"""
        idx = node.tokens.find_fuzzy(
            ["g_task_return_new_error", "(", "~*", ",", "~*", ",", "~*", ",", "~*", ")"]
        )
        if idx != -1:
            token = node.tokens[idx]
            self.add_failure(
                "missing literal, use g_task_return_new_error_literal() instead",
                linecnt=token.linecnt,
            )
        idx = node.tokens.find_fuzzy(["g_prefix_error", "(", "~*", ",", "~*", ")"])
        if idx != -1:
            token = node.tokens[idx]
            self.add_failure(
                "missing literal, use g_prefix_error_literal() instead",
                linecnt=token.linecnt,
            )

        idx = node.tokens.find_fuzzy(
            ["g_set_error", "(", "~*", ",", "~*", ",", "~*", ",", "~*", ")"]
        )
        if idx != -1:
            if node.tokens[idx + 8].data.find("%m") == -1:
                self.add_failure(
                    "missing literal, use g_set_error_literal() instead",
                    linecnt=node.linecnt,
                )

    def _test_missing_error_suffixes(self, node: Node) -> None:
        """test for missing : suffixes"""
        idx = node.tokens.find_fuzzy(["~g_prefix_error*", "(", "~*", ","])
        if idx == -1:
            return
        token = node.tokens[idx + 4]
        if not token.data.endswith(': "'):
            self.add_failure("missing ': ' suffix", linecnt=token.linecnt)

    def _test_extra_error_suffixes(self, node: Node) -> None:
        """test for extra : suffixes"""
        idx = node.tokens.find_fuzzy(["~g_set_error*", "(", "~*[", ","])
        if idx == -1:
            return
        token = node.tokens[idx + 8]
        if token.data.endswith(': "'):
            self.add_failure("extraneous ': ' suffix", linecnt=token.linecnt)

    def _test_enums(self, node: Node) -> None:
        if node.depth != 0:
            return

        # only consider the last token
        idx = node.tokens_pre.find_fuzzy(["@ENUM"], offset=len(node.tokens_pre) - 1)
        if idx == -1:
            return
        name = node.tokens_pre[idx].data
        if self.verbose:
            print("enum_name", name)

        # needs Fu prefix
        valid_prefixes = ["Fwupd", "Fu"]
        for prefix in valid_prefixes:
            if name.startswith(prefix):
                return
        self.add_failure(
            f"invalid enum name {name} should have {'|'.join(valid_prefixes)} prefix",
            linecnt=node.linecnt,
        )

    def _test_struct(self, node: Node) -> None:

        if node.depth != 0:
            return

        # no limit on name
        if node.hint == NodeHint.STRUCT_CONST:
            return

        # only consider the last token
        name: Optional[str] = None
        if node.hint == NodeHint.STRUCT_TYPEDEF:
            name = node.tokens_pre[-1].data
        else:
            idx = node.tokens_pre.find_fuzzy(
                ["@STRUCT"], offset=len(node.tokens_pre) - 1
            )
            if idx != -1:
                name = node.tokens_pre[idx].data
        if not name:
            return
        if name.startswith("_"):
            name = name[1:]
        if self.verbose:
            print("struct_name", name)

        # only for Rust code
        if name.startswith("FuStruct"):
            self.add_failure(
                f"incorrect struct name {name} -- FuStruct is reserved for Rust code",
                linecnt=node.linecnt,
            )

        # needs Fu prefix
        valid_prefixes = ["Fwupd", "Fu"]
        for prefix in valid_prefixes:
            if name.startswith(prefix):
                return
        self.add_failure(
            f"invalid struct name {name} should have {'|'.join(valid_prefixes)} prefix",
            linecnt=node.linecnt,
        )

    def _test_magic_numbers_buffer(self, node: Node) -> None:

        # skip tests
        if self._current_fn:
            basename = os.path.basename(self._current_fn)
            if basename.endswith("-test.c"):
                return

        limit: int = 4
        idx = node.tokens.find_fuzzy(["~*@INTEGER", ","] * limit)
        if idx == -1:
            return
        token = node.tokens[idx]
        self.add_failure(
            f"variable has too many magic values (limit of {limit})",
            linecnt=token.linecnt,
        )

    def _test_static_vars(self, node: Node) -> None:

        if node.depth != 0:
            return

        idx = node.tokens_pre.find_fuzzy(["static", "~*", "~*", "="])
        if idx == -1:
            return

        token = node.tokens_pre[idx + 2]
        if token.data in ["signals", "quarks"]:
            return
        self.add_failure(
            f"static variable {token.data} not allowed", linecnt=token.linecnt
        )

    def _test_rustgen_bitshifts(self, node: Node) -> None:

        if node.hint == NodeHint.ENUM:
            return
        idx = node.tokens.find_fuzzy(["]", "<", "<", "16"])
        if idx != -1 and node.tokens.find_fuzzy(["]", "<", "<", "8"]) != -1:
            token = node.tokens[idx + 3]
            self.add_failure(
                "endian unsafe construction; perhaps use fu_memread_uint32 or rustgen",
                linecnt=token.linecnt,
            )

    def _test_rustgen_vars(self, node: Node) -> None:

        idx = node.tokens.find_fuzzy(["g_autoptr", "(", "~FuStruct*", ")", "~*", "="])
        if idx == -1:
            return
        token = node.tokens[idx + 4]
        if not token.data.startswith("st"):
            self.add_failure(
                f"rustgen structure '{token.data}' has to have 'st' prefix",
                linecnt=token.linecnt,
            )

    def _test_zero_init(self, node: Node) -> None:
        if node.hint in [
            NodeHint.UNION,
            NodeHint.STRUCT,
            NodeHint.STRUCT_CONST,
            NodeHint.STRUCT_TYPEDEF,
        ]:
            return
        if node.tokens_pre.count_fuzzy(["struct"]) > 0:
            return
        idx = node.tokens.find_fuzzy(["~guint*", "~*", "[", "~*", "]", ";"])
        if idx != -1:
            token = node.tokens[idx]
            self.add_failure(
                "buffer not zero init, use ` = {0}`", linecnt=token.linecnt
            )

    def _test_debug_newlines(self, node: Node) -> None:

        for func in ["g_info", "g_debug", "g_message"]:
            idx: int = 0
            while True:
                idx = node.tokens.find_fuzzy([func, "(", "~*\\n*"], offset=idx + 1)
                if idx == -1:
                    break
                token = node.tokens[idx]
                self.add_failure(
                    f"{func} should not contain newlines", linecnt=token.linecnt
                )

    def _test_debug_fullstops(self, node: Node) -> None:

        for func in ["g_error", "g_info", "g_debug", "g_message"]:
            idx: int = 0
            while True:
                idx = node.tokens.find_fuzzy([func, "(", '~*."'], offset=idx + 1)
                if idx == -1:
                    break
                token = node.tokens[idx]
                self.add_failure(
                    f"{func} should not end with a full stop", linecnt=token.linecnt
                )

    def _test_debug_sentence_case(self, node: Node) -> None:

        for func in ["g_error", "g_info", "g_debug", "g_message"]:
            idx: int = 0
            while True:
                idx = node.tokens.find_fuzzy([func], offset=idx + 1)
                if idx == -1:
                    break
                token = node.tokens[idx + 2]
                first_word = token.data[1:].split(" ")[0]
                if first_word and first_word[0].isupper() and first_word[1:].islower():
                    self.add_failure(
                        f"{func} should not use sentence case", linecnt=token.linecnt
                    )

    def _test_comment_cpp(self, node: Node) -> None:
        """/* C comments please */"""
        idx = node.tokens.find_fuzzy(["//"])
        if idx != -1:
            token = node.tokens[idx]
            self.add_failure(
                f"use C style comments, not C++, e.g. /* this */",
                linecnt=token.linecnt,
            )

    def _test_comment_lines(self, node: Node) -> None:
        """/**** not boxes *****/"""
        for token in node.tokens_pre + node.tokens:
            if token.data.find("/***") != -1:
                self.add_failure(
                    f"do not use boxed comment lines, just use /* comment */",
                    linecnt=token.linecnt,
                )
            if token.data.find("/**@brief") != -1:
                self.add_failure(
                    f"do not use doxygen comment lines, just use /* comment */",
                    linecnt=token.linecnt,
                )

    def _test_comment_lower_case(self, node: Node) -> None:
        """single line comments are supposed to be lowercase"""
        idx: int = 0
        while True:
            idx = node.tokens.find_fuzzy(["@COMMENT"], offset=idx + 1)
            if idx == -1:
                break
            token = node.tokens[idx]
            first_word = token.data[2:-2].strip().split(" ")[0]
            if (
                first_word not in ["Windows", "Microsoft", "Thunderbolt", "Dell"]
                and token.linecnt == token.linecnt_end
                and first_word
                and first_word[0].isupper()
                and first_word[1:].islower()
            ):
                self.add_failure(
                    f"single line comments should not use sentence case",
                    linecnt=token.linecnt,
                )

    def _test_equals_true(self, node: Node) -> None:
        """do not allow `if (foo == TRUE)`"""

        idx = node.tokens_pre.find_fuzzy(["==", "TRUE", ")"])
        if idx != -1:
            token = node.tokens_pre[idx]
            self.add_failure(
                f"do not compare a boolean to TRUE",
                linecnt=token.linecnt,
            )

    def _test_blocked_goto(self, node: Node) -> None:
        idx = node.tokens.find_fuzzy(["goto"])
        if idx != -1:
            token = node.tokens_pre[idx]
            self.add_failure(
                f"do not use goto, refactor into a new block",
                linecnt=token.linecnt,
            )

    def _test_device_display(self, node: Node) -> None:
        """use fu_device_get_id_display rather than the two different commands"""

        idx1 = node.tokens.find_fuzzy(["fu_device_get_name", "(", "~*", ")"])
        idx2 = node.tokens.find_fuzzy(["fu_device_get_id", "(", "~*", ")"])
        if idx1 != -1 and idx2 != -1:

            # check this isn't an assert
            if node.tokens[idx1 - 2].data == "g_assert_cmpstr":
                return

            # same FuDevice within a limited number of lines?
            token1 = node.tokens[idx1 + 2]
            token2 = node.tokens[idx2 + 2]
            limit: int = 3
            if token1.data == token2.data:
                if abs(token1.linecnt - token2.linecnt) < limit:
                    self.add_failure(
                        "use fu_device_get_id_display() rather than "
                        "fu_device_get_name()+fu_device_get_id()",
                        linecnt=token1.linecnt,
                    )

    def _test_debug_fns(self, node: Node) -> None:
        # no console output expected
        if self._current_fn and os.path.basename(self._current_fn) in [
            "fu-console.c",
            "fu-daemon.c",
            "fu-dbxtool.c",
            "fu-debug.c",
            "fu-fuzzer-main.c",
            "fu-gcab.c",
            "fu-main.c",
            "fu-main-windows.c",
            "fu-self-test.c",
            "fu-tpm-eventlog.c",
            "fwupd-self-test.c",
        ]:
            return
        for token, msg in {
            "~g_print*": "Use g_debug() instead",
        }.items():
            idx = node.tokens.find_fuzzy([token])
            if idx != -1:
                token = node.tokens[idx]
                self.add_failure(
                    f"contains blocked token {token.data}: {msg}", linecnt=token.linecnt
                )

    def _test_gobject_finalize(self, node: Node) -> None:

        if node.tokens_pre.endswith_fuzzy(
            ["void", "~*_finalize", "(", "GObject", "*", "~*", ")"]
        ):
            token = node.tokens_pre[-1]
            idx = node.tokens.find_fuzzy(
                ["G_OBJECT_CLASS", "(", "~*_parent_class", ")", "-", ">", "finalize"]
            )
            if idx == -1:
                self.add_failure(
                    "did not have parent ->finalize()", linecnt=token.linecnt
                )

    def _test_gobject_constructed(self, node: Node) -> None:

        if node.tokens_pre.endswith_fuzzy(
            ["void", "~*_constructed", "(", "GObject", "*", "~*", ")"]
        ):
            token = node.tokens_pre[-1]
            idx = node.tokens.find_fuzzy(
                ["G_OBJECT_CLASS", "(", "~*_parent_class", ")", "-", ">", "constructed"]
            )
            if idx == -1:
                self.add_failure(
                    "did not have parent ->constructed()", linecnt=token.linecnt
                )

    def _test_blocked_funcs(self, node: Node) -> None:

        for token, msg in {
            "g_error": "Use GError instead",
            "g_ascii_strtoull": "Use fu_strtoull() instead",
            "g_ascii_strtoll": "Use fu_strtoll() instead",
            "g_find_program_in_path": "Use fu_path_store_find_program() instead",
            "g_strerror": "Use fwupd_strerror() instead",
            "gnutls_malloc": "Use gnutls_calloc() instead",
            "g_base64_encode": "Use fu_base64_encode() instead",
            "g_variant_get_uint64": "Use fwupd_variant_get_uint64() instead",
            "g_variant_get_uint32": "Use fwupd_variant_get_uint32() instead",
            "g_variant_get_string": "Use fwupd_variant_get_string() instead",
            "g_variant_get_boolean": "Use fwupd_variant_get_boolean() instead",
            "g_variant_get_strv": "Use fwupd_variant_get_strv() instead",
            "g_variant_get_double": "Use fwupd_variant_get_double() instead",
            "g_variant_get_int32": "Use fwupd_variant_get_int32() instead",
            "g_random_int_range": "Use fu_common_get_random() or a predictable token instead",
            "g_assert": "Use g_set_error() or g_return_val_if_fail() instead",
            "HIDIOCSFEATURE": "Use fu_hidraw_device_set_feature() instead",
            "HIDIOCGFEATURE": "Use fu_hidraw_device_get_feature() instead",
            "memcpy": "Use fu_memcpy_safe or rustgen instead",
            "g_unsetenv": "This is not thread safe",
            "~GUINT??_FROM_?E": "Use fu_memread_uintXX_safe() or rustgen instead",
            "~GUINT??_TO_?E": "Use fu_memwrite_uintXX_safe() or rustgen instead",
            "ioctl": "Use fu_udev_device_ioctl() instead",
        }.items():
            idx = node.tokens.find_fuzzy([token, "("])
            if idx != -1:
                token = node.tokens[idx]
                self.add_failure(
                    f"contains blocked token {token.data}: {msg}", linecnt=token.linecnt
                )

        idx = node.tokens.find_fuzzy(["|=", "~1*", "<", "<"])
        if idx != -1:
            token = node.tokens[idx]
            self.add_failure("Use FU_BIT_SET() instead", linecnt=token.linecnt)
        idx = node.tokens.find_fuzzy(["|=", "(", "~1*", "<", "<"])
        if idx != -1:
            token = node.tokens[idx]
            self.add_failure("Use FU_BIT_SET() instead", linecnt=token.linecnt)
        idx = node.tokens.find_fuzzy(["&", "=", "~", "(", "~1*", "<", "<"])
        if idx != -1:
            token = node.tokens[idx]
            self.add_failure("Use FU_BIT_CLEAR() instead", linecnt=token.linecnt)
        idx = node.tokens_pre.find_fuzzy(
            ["__attribute__", "(", "(", "packed", ")", ")"]
        )
        if idx != -1:
            token = node.tokens_pre[idx]
            self.add_failure("use rustgen instead", linecnt=token.linecnt)

    def _test_blocked_tokens(self, node: Node) -> None:

        for search, msg in {
            "__FUNCTION__": "Use G_STRFUNC instead",
            "__VA_ARGS__": "Use native functions instead",
            "uint32_t": "Use guint32 instead",
            "uint16_t": "Use guint16 instead",
            "uint8_t": "Use guint8 instead",
        }.items():
            for token in node.tokens:
                if token.data.find(search) != -1:
                    self.add_failure(
                        f"contains blocked token {token.data}: {msg}",
                        linecnt=token.linecnt,
                    )

    def _test_magic_numbers_defined(self, nodes: List[Node]) -> None:

        cnt: int = 0
        limit: int = 15
        linecnt: int = 0
        for node in nodes:
            if node.depth != 0:
                continue

            # overridden per-file
            idx = node.tokens_pre.find_fuzzy([f"~*{self._current_nocheck}=*"])
            if idx != -1:
                limit = int(node.tokens_pre[idx].data.split("=")[1].split(" ")[0])
            cnt_tmp = node.tokens_pre.count_fuzzy(["#define", "~*", "~0x*"])
            if cnt_tmp:
                linecnt = node.linecnt
                cnt += cnt_tmp
        if cnt > limit:
            self.add_failure(
                f"file has too many #defined magic values ({cnt}), limit of {limit}",
                linecnt=linecnt,
            )

    def _test_magic_numbers_inline(self, nodes: List[Node]) -> None:

        cnt: int = 0
        limit: int = 80
        linecnt: int = 0
        for node in nodes:
            # overridden per-file
            idx = node.tokens_pre.find_fuzzy([f"~*{self._current_nocheck}=*"])
            if idx != -1:
                for word in node.tokens_pre[idx].data.split(" "):
                    if word.startswith("nocheck:magic-inlines="):
                        limit = int(word[22:])
                        break
            for token in node.tokens:
                if token.data in ["0x0", "0x00", "0x0000"]:
                    continue
                if len(token.data) >= 3 and token.data.startswith("0x"):
                    linecnt = node.linecnt
                    cnt += 1
        if cnt > limit:
            self.add_failure(
                f"file has too many inline magic values ({cnt}), limit of {limit}",
                linecnt=linecnt,
            )

    def _test_gerror_false_returns(self, nodes: List[Node]) -> None:

        for node in nodes:
            if node.depth == 0:
                continue
            if not self._should_process_node(node):
                return
            idx = node.tokens.find_fuzzy(["~g_set_error*", "("])
            if idx != -1:
                token = node.tokens[idx]
                idx = node.tokens.find_fuzzy(["return", "~*", ";"], offset=idx)
                if idx == -1:
                    idx = node.tokens.find_fuzzy(["break", ";"], offset=idx)
                if idx == -1:
                    self.add_failure(
                        "uses g_set_error() without returning a value",
                        linecnt=token.linecnt,
                    )
                    break

    def _test_gerror_not_set(self, nodes: List[Node]) -> None:

        limit: int = 10
        for node in nodes:
            if node.depth == 0:
                continue
            if not self._should_process_node(node):
                continue
            idx = node.tokens.find_fuzzy(["~g_prefix_error*", "("])
            if idx != -1:
                linecnt = node.tokens[idx].linecnt
                if self.verbose:
                    print(f"GError required @{linecnt}")

                found_linecnt: List[int] = []

                # set error inner
                idx_found = node.tokens.find_fuzzy(
                    ["~error*", ")"], reverse=True, offset=idx
                )
                if idx_found != -1:
                    found_linecnt.append(node.tokens[idx_found].linecnt_end)
                idx_found = node.tokens.find_fuzzy(
                    ["~g_set_error*", "(", "~error*"], reverse=True, offset=idx
                )
                if idx_found != -1:
                    found_linecnt.append(node.tokens[idx_found].linecnt_end)

                # set error prior
                idx_found = node.tokens_pre.find_fuzzy(["~error*", ")"], reverse=True)
                if idx_found != -1:
                    found_linecnt.append(node.tokens_pre[idx_found].linecnt_end)
                idx_found = node.tokens_pre.find_fuzzy(
                    ["~g_set_error*", "(", "~error*"], reverse=True
                )
                if idx_found != -1:
                    found_linecnt.append(node.tokens_pre[idx_found].linecnt_end)
                idx_found = node.tokens_pre.find_fuzzy(["~error*", ","], reverse=True)
                if idx_found != -1:
                    found_linecnt.append(node.tokens_pre[idx_found].linecnt_end)

                # find the closest error set
                linecnt_closest: int = 0
                if found_linecnt:
                    linecnt_closest = max(found_linecnt)
                    if self.verbose:
                        print(f"GError set @{linecnt_closest}")

                linecnt_delta = linecnt - linecnt_closest
                if linecnt_delta > limit:
                    self.add_failure(
                        "uses g_prefix_error() without setting GError for "
                        f"{linecnt_delta}/{limit} previous lines",
                        linecnt=linecnt,
                    )
                    break

    def _test_gerror_domain(self, node: Node) -> None:
        """must use FUWPD_ERROR domains"""
        idx = node.tokens.find_fuzzy(
            ["~g_set_error*", "(", "~*", ",", "~G_*", ",", "~G_*"]
        )
        if idx == -1:
            return
        token = node.tokens[idx]
        self.add_failure(
            "uses g_set_error() without using FWUPD_ERROR", linecnt=token.linecnt
        )

    def _test_gerror_deref(self, node: Node) -> None:
        """using (*error)->message"""
        idx = node.tokens.find_fuzzy(["(", "*", "~*error*", ")", "-", ">"])
        if idx == -1:
            return
        token = node.tokens[idx]
        self.add_failure(
            "dereferences GError; use error_local instead", linecnt=token.linecnt
        )

    def _test_switch(self, nodes: List[Node]) -> None:

        limit: int = 2
        cnt: int = 0
        for node in nodes:
            # restrict only per-top-level=function, not per-file
            if node.depth == 0:
                cnt = 0
            idx = node.tokens_pre.find_fuzzy(["switch", "(", "~*", ")"])
            if idx != -1:
                cnt += 1
                if cnt > limit:
                    self.add_failure(
                        f"has too many switches ({cnt}), limit of {limit}",
                        linecnt=node.linecnt,
                    )
                    break

    def _test_null_false_returns(self, nodes: List[Node]) -> None:

        # allowed values from g_return_val_if_fail()
        types_rvif = {
            "*": ["NULL"],
            "GQuark": ["0"],
            "GType": ["G_TYPE_INVALID"],
            "gpointer": ["NULL"],
            "gboolean": ["TRUE", "FALSE"],
            "guint32": ["0", "G_MAXUINT32"],
            "guint64": ["0", "G_MAXUINT64"],
            "guint16": ["0", "G_MAXUINT16"],
            "guint8": ["0", "G_MAXUINT8"],
            "gint64": ["0", "-1", "G_MAXINT64"],
            "gint32": ["0", "-1", "G_MAXINT32"],
            "gint16": ["0", "-1", "G_MAXINT16"],
            "gint8": ["0", "-1", "G_MAXINT8"],
            "gint": ["0", "-1", "G_MAXINT"],
            "guint": ["0", "G_MAXUINT"],
            "gulong": ["0", "G_MAXLONG"],
            "gsize": ["0", "G_MAXSIZE"],
            "gssize": ["0", "-1", "G_MAXSSIZE"],
        }

        # disallowed values from return
        types_nret = {
            "*": ["FALSE"],
            "GQuark": ["NULL", "TRUE", "FALSE"],
            "GType": ["NULL", "TRUE", "FALSE"],
            "gpointer": ["FALSE"],
            "gboolean": ["NULL", "0"],
            "guint32": ["NULL", "TRUE", "FALSE"],
            "guint64": ["NULL", "TRUE", "FALSE"],
            "guint16": ["NULL", "TRUE", "FALSE"],
            "guint8": ["NULL", "TRUE", "FALSE"],
            "gint64": ["NULL", "TRUE", "FALSE"],
            "gint32": ["NULL", "TRUE", "FALSE"],
            "gint16": ["NULL", "TRUE", "FALSE"],
            "gint8": ["NULL", "TRUE", "FALSE"],
            "gint": ["NULL", "TRUE", "FALSE"],
            "guint": ["NULL", "TRUE", "FALSE"],
            "gulong": ["NULL", "TRUE", "FALSE"],
            "gsize": ["NULL", "TRUE", "FALSE"],
            "gssize": ["NULL", "TRUE", "FALSE"],
        }

        current_type: str = ""
        for node in nodes:
            if not self._should_process_node(node):
                continue

            # new function
            if node.depth == 0:
                idx = node.tokens_pre.find_fuzzy(["~*", "("])
                if idx == -1:
                    continue
                token_type = node.tokens_pre[idx - 1]
                token_func = node.tokens_pre[idx]
                if self.verbose:
                    print("TYPE", token_func.data, token_type.data)
                current_type = token_type.data

            # look for g_return_val_if_fail
            idx = node.tokens.find_fuzzy(["g_return_val_if_fail", "("])
            if idx != -1:
                # advance to the return
                idx = node.tokens.find_fuzzy([",", "~*", ")", ";"], offset=idx)
                if idx != -1:
                    token = node.tokens[idx + 1]
                    try:
                        rvif = types_rvif[current_type]
                    except KeyError:
                        if self.verbose:
                            print(f"missing type {current_type}")
                        continue
                    if _value_relaxed(token.data) not in rvif:
                        self.add_failure(
                            "g_return_val_if_fail() return type invalid, "
                            f"expected {'|'.join(rvif)} for {current_type}, not {token.data}",
                            linecnt=token.linecnt,
                        )

            # look for return values
            idx = node.tokens.find_fuzzy(["return", "~*", ";"])
            if idx == -1:
                continue
            token = node.tokens[idx + 1]
            if token.hint == TokenHint.FUNCTION:
                continue
            try:
                nret = types_nret[current_type]
            except KeyError:
                continue
            if _value_relaxed(token.data) in nret:
                self.add_failure(
                    f"return type invalid for {current_type}: {token.data}",
                    linecnt=token.linecnt,
                )

    def _test_gerror_void_return(self, node: Node) -> None:
        """takes GError but returns void"""

        if node.depth != 0:
            return

        idx_start = node.tokens_pre.find_fuzzy(["static", "void", "@FUNCTION"])
        if idx_start == -1:
            return
        idx = node.tokens_pre.find_fuzzy(
            ["GError", "*", "*", "error", ")"], offset=idx_start
        )
        if idx == -1:
            return
        if idx - idx_start < 10:
            token = node.tokens_pre[idx]
            self.add_failure(
                "void return type not expected for GError",
                linecnt=token.linecnt,
            )

    def _test_function_length(self, node: Node) -> None:

        if node.depth != 0:
            return
        limit: int = 400
        if node.linecnt_end - node.linecnt > limit:
            self.add_failure(
                f"function is too long, was {node.linecnt_end - node.linecnt} of {limit}",
                linecnt=node.linecnt,
            )

    def _test_firmware_convert_version(self, nodes: List[Node]) -> None:

        # contains fu_firmware_set_version_raw()
        _set_version_raw: bool = False
        for node in nodes:
            idx = node.tokens.find_fuzzy(
                ["fu_firmware_set_version_raw", "(", "~*", ",", "~*", ")", ";"]
            )
            if idx != -1:
                _set_version_raw = True
        if not _set_version_raw:
            return

        # also contains fu_firmware_set_version()
        for node in nodes:
            if not self._should_process_node(node):
                continue
            idx = node.tokens.find_fuzzy(
                ["fu_firmware_set_version", "(", "~*", ",", "~*", ")", ";"]
            )
            if idx != -1:
                token = node.tokens[idx]
                self.add_failure(
                    "Use FuFirmwareClass->convert_version rather than fu_firmware_set_version()",
                    linecnt=token.linecnt,
                )

    def _test_device_convert_version(self, nodes: List[Node]) -> None:

        if self._current_fn and os.path.basename(self._current_fn) in [
            "fu-engine-test.c",
        ]:
            return

        # contains fu_device_set_version_raw()
        _set_version_raw: bool = False
        for node in nodes:
            idx = node.tokens.find_fuzzy(
                ["fu_device_set_version_raw", "(", "~*", ",", "~*", ")", ";"]
            )
            if idx != -1:
                _set_version_raw = True
        if not _set_version_raw:
            return

        # also contains fu_device_set_version()
        for node in nodes:
            if not self._should_process_node(node):
                continue
            idx = node.tokens.find_fuzzy(
                ["fu_device_set_version", "(", "~*", ",", "~*", ")", ";"]
            )
            if idx != -1:
                token = node.tokens[idx]
                self.add_failure(
                    "Use FuDeviceClass->convert_version rather than fu_device_set_version()",
                    linecnt=token.linecnt,
                )

    def _test_small_conditionals_with_braces_node(self, node: Node) -> None:

        # has previous conditional
        idx = node.tokens_pre.find_fuzzy(["else", "if", "("])
        if idx != -1:
            return

        # has conditional
        idx = node.tokens_pre.find_fuzzy(["if", "("])
        if idx == -1:
            return

        # multiline conditional can have braces
        if node.linecnt_end - node.tokens_pre[idx].linecnt >= 3:
            return

        # inline struct
        idx = node.tokens_pre.find_fuzzy(["@STRUCT"])
        if idx != -1:
            return

        # nested
        idx = node.tokens.find_fuzzy(["if", "("])
        if idx != -1:
            return

        # not one line block
        if len(node.tokens) < 2:
            return
        if node.tokens[1].linecnt != node.tokens[-1].linecnt:
            return
        limit: int = 70
        if len("".join([token.data for token in node.tokens])) < limit:
            token = node.tokens[1]
            self.add_failure(
                "no {} required for small single-line conditional",
                linecnt=token.linecnt,
            )

    def _test_small_conditionals_with_braces(self, nodes: List[Node]) -> None:

        # we need to parse the nodes in order
        for idx, node in enumerate(nodes):
            next_node_depth: int = 0
            try:
                next_node_depth = nodes[idx + 1].depth
            except IndexError:
                pass

            # followed by else
            try:
                idx = nodes[idx + 1].tokens_pre.find_fuzzy(["else"])
                if idx != -1:
                    continue
            except IndexError:
                pass

            # ignore if is nested deeper
            if next_node_depth > node.depth:
                continue

            # check node
            self._test_small_conditionals_with_braces_node(node)

    def _test_nesting_depth(self, node: Node) -> None:

        limit: int = 5
        if node.depth >= limit:
            self.add_failure(
                f"is nested too deep {node.depth}/{limit}", linecnt=node.linecnt
            )

    def _test_strsafe_offset(self, node: Node) -> None:

        idx = node.tokens.find_fuzzy(
            [
                "fu_strsafe@FUNCTION",
                "(",
                "(",
                "const",
                "~*char",
                "*",
                ")",
                "~*",
                "+",
            ]
        )
        if idx == -1:
            idx = node.tokens.find_fuzzy(
                [
                    "fu_strsafe@FUNCTION",
                    "(",
                    "(",
                    "gchar",
                    "*",
                    ")",
                    "~*",
                    "+",
                ]
            )
        if idx == -1:
            idx = node.tokens.find_fuzzy(
                [
                    "fu_strsafe@FUNCTION",
                    "(",
                    "(",
                    "const",
                    "gchar",
                    "*",
                    ")",
                    "~*",
                    "-",
                    ">",
                    "data",
                    "+",
                ]
            )
        if idx != -1:
            token = node.tokens[idx]
            self.add_failure(
                "use fu_memstrsafe() rather than fu_strsafe() if reading with an offset",
                linecnt=token.linecnt,
            )

    def _test_memread(self, node: Node) -> None:

        limit: int = 7
        cnt = node.tokens.count_fuzzy(["~fu_memread_uint*"])
        if cnt >= limit:
            self.add_failure(
                f"too many calls to fu_memread_uintXX() ({cnt}/{limit}), use rustgen",
                linecnt=node.linecnt,
            )
        cnt = node.tokens.count_fuzzy(["~fu_memwrite_uint*"])
        if cnt >= limit:
            self.add_failure(
                f"too many calls to fu_memwrite_uintXX() ({cnt}/{limit}), use rustgen",
                linecnt=node.linecnt,
            )

    def _test_gobject_parents(self, nodes: List[Node]) -> None:

        gtype: str = ""
        gtypeparent: str = ""
        for node in nodes:
            if node.depth != 0:
                continue
            if not self._should_process_node(node):
                return

            # .h
            idx = node.tokens_pre.find_fuzzy(["~G_DECLARE_*_TYPE", "("])
            if idx != -1:
                gtype = node.tokens_pre[idx + 2].data
                gtypeparent = node.tokens_pre[idx + 10].data
                self._gtype_parents[gtype] = gtypeparent

            # check the class def is correct
            if gtype:
                idx = node.tokens.find_fuzzy(["~*Class", "parent_class"])
                if idx != -1:
                    gtypeparentclass_found = node.tokens[idx].data
                    gtypeparentclass_expected = f"{gtypeparent}Class"
                    if gtypeparentclass_found != gtypeparentclass_expected:
                        self.add_failure(
                            f"wrong parent_class for {gtype}, "
                            f"got {gtypeparentclass_found} and "
                            f"expected {gtypeparentclass_expected}",
                            linecnt=node.linecnt,
                        )

            # .c
            idx = node.tokens_pre.find_fuzzy(["~G_DEFINE_TYPE*", "("])
            if idx != -1:
                gtype = node.tokens_pre[idx + 2].data
                if self.verbose:
                    print("GTYPE", gtype, self._gtype_parents)
                if not gtype in self._gtype_parents:
                    continue
                gtypeparent_found: str = node.tokens_pre[idx + 6].data
                gtype_snake = _camel_to_snake(self._gtype_parents[gtype]).split(
                    "_", maxsplit=1
                )
                gtypeparent_expected = f"{gtype_snake[0]}_type_{gtype_snake[1]}".upper()
                if gtypeparent_found != gtypeparent_expected:
                    self.add_failure(
                        f"wrong parent GType for {gtype}, "
                        f"got {gtypeparent_found} and "
                        f"expected {gtypeparent_expected}",
                        linecnt=node.linecnt,
                    )

    def _test_nodes(self, nodes: List[Node]) -> None:

        # preroll
        self._klass_funcs.clear()
        for node in nodes:
            self._discover_klass_functions(node)
        if self.verbose:
            print("KLASS FUNCS", self._klass_funcs)

        # tests we can do node by node
        for node in nodes:

            # test for missing ->finalize()
            self._current_nocheck = "nocheck:finalize"
            if self._should_process_node(node):
                self._test_gobject_finalize(node)
                self._test_gobject_constructed(node)

            # test for blocked functions
            self._current_nocheck = "nocheck:blocked"
            if self._should_process_node(node):
                self._test_blocked_funcs(node)
                self._test_blocked_tokens(node)
                self._test_blocked_goto(node)
                self._test_device_display(node)
                self._test_equals_true(node)

            # test for debug lines
            self._current_nocheck = "nocheck:print"
            if self._should_process_node(node):
                self._test_debug_fns(node)
                self._test_debug_newlines(node)
                self._test_debug_fullstops(node)
                self._test_debug_sentence_case(node)
                self._test_comment_lower_case(node)
                self._test_comment_cpp(node)
                self._test_comment_lines(node)

            # not nesting too deep
            self._current_nocheck = "nocheck:depth"
            if self._should_process_node(node):
                self._test_nesting_depth(node)

            # test for function names
            self._current_nocheck = "nocheck:name"
            if self._should_process_node(node):
                self._test_function_names_prefix(node)
                self._test_function_names_ensure(node)
                self._test_function_name_suffix(node)
                self._test_param_self_device(node)
                self._test_param_self_firmware(node)
                self._test_param_self_native_device(node)
                self._test_param_self_native_firmware(node)
                self._test_variable_case(node)
                self._test_variable_buffer_index(node)
                self._test_struct_member_case(node)

            # test for invalid struct and enum names
            self._current_nocheck = "nocheck:prefix"
            if self._should_process_node(node):
                self._test_struct(node)
                self._test_enums(node)

            # test for static variables
            self._current_nocheck = "nocheck:static"
            if self._should_process_node(node):
                self._test_static_vars(node)

            # test for magic numbers
            self._current_nocheck = "nocheck:magic"
            if self._should_process_node(node):
                self._test_magic_numbers_buffer(node)

            # test for rustgen variables
            self._current_nocheck = "nocheck:rustgen"
            if self._should_process_node(node):
                self._test_rustgen_vars(node)

            # test for endian safety
            self._current_nocheck = "nocheck:endian"
            if self._should_process_node(node):
                self._test_rustgen_bitshifts(node)

            # GError
            self._current_nocheck = "nocheck:error"
            if self._should_process_node(node):
                self._test_gerror_domain(node)
                self._test_gerror_void_return(node)
                self._test_extra_error_suffixes(node)
                self._test_missing_error_suffixes(node)
                self._test_missing_literal(node)
                self._test_gerror_deref(node)

            # using too many memory reads/writes
            self._current_nocheck = "nocheck:memread"
            if self._should_process_node(node):
                self._test_memread(node)
                self._test_strsafe_offset(node)

            # test for non-zero'd init
            self._current_nocheck = "nocheck:zero-init"
            if self._should_process_node(node):
                self._test_zero_init(node)

            # functions too long
            self._current_nocheck = "nocheck:lines"
            if self._should_process_node(node):
                self._test_function_length(node)

        # NULL != FALSE
        self._current_nocheck = "nocheck:lines"
        self._test_null_false_returns(nodes)
        self._test_small_conditionals_with_braces(nodes)

        # using too many switch()
        self._current_nocheck = None
        self._test_switch(nodes)

        # should use FuFirmwareClass->convert_version or FuDeviceClass->convert_version
        self._current_nocheck = "nocheck:set-version"
        self._test_firmware_convert_version(nodes)
        self._test_device_convert_version(nodes)

        # using too many hardcoded constants
        self._current_nocheck = "nocheck:magic-defines"
        self._test_magic_numbers_defined(nodes)
        self._current_nocheck = "nocheck:magic-inlines"
        self._test_magic_numbers_inline(nodes)

        # setting GError, not returning
        self._current_nocheck = "nocheck:error-false-return"
        self._test_gerror_false_returns(nodes)

        # prefix with no set
        self._current_nocheck = "nocheck:error"
        self._test_gerror_not_set(nodes)

        # test GObject parent types
        self._current_nocheck = "nocheck:name"
        self._test_gobject_parents(nodes)

    def test_file(self, fn: str) -> None:
        self._current_fn = fn
        self._current_nocheck = None
        with open(fn, "rb") as f:
            try:
                data = f.read().decode()
            except UnicodeDecodeError as e:
                print(f"failed to read {fn}: {e}")
        if data.find("Copyright") == -1:
            self.add_failure(f"does not have copyright assigned")
        if data.find("Copyright") == -1:
            self.add_failure(f"does not have a SPDX-License-Identifier")
        tokenizer = Tokenizer(data)
        nodes = tokenizer.nodes
        if self.verbose:
            print(nodes)
        self._test_nodes(nodes)


def test_files(fns_optional: List[str], verbose: bool = False) -> int:
    # test all C and H files

    checker = Checker(verbose=verbose)

    # use any file specified in argv, falling back to scanning the entire tree
    fns: List[str] = []
    if fns_optional:
        for fn in fns_optional:
            if fn.startswith("contrib/ci/tests"):
                continue
            try:
                ext: str = fn.rsplit(".", maxsplit=1)[1]
            except IndexError:
                continue
            if ext in ["c", "h"]:
                fns.append(fn)
    else:
        fns.extend(glob.glob("libfwupd/*.[c|h]"))
        fns.extend(glob.glob("libfwupdplugin/*.[c|h]"))
        fns.extend(glob.glob("plugins/*/*.[c|h]"))
        fns.extend(glob.glob("src/*.[c|h]"))
    for fn in sorted(fns, reverse=True):
        if os.path.basename(fn) == "check-source.py":
            continue
        print(f"checking {fn}…", file=sys.stderr)
        checker.test_file(fn)

    # show issues
    for failure in checker.failures:
        line: str = ""
        if failure.fn:
            line += failure.fn
        if failure.linecnt:
            line += f":{failure.linecnt}"
        line += f": {failure.message}"
        if failure.nocheck:
            line += f" -- use a {failure.nocheck} comment to ignore"
        print(line)
    if checker.failures:
        print(f"{len(checker.failures)} failures!")
    return 1 if checker.failures else 0


def unit_test(fn: str, verbose: bool = False) -> int:

    # load test file with any expected failures
    rc: int = 0
    checker = Checker(verbose=verbose)
    with open(fn, "rb") as f:
        lines = f.read().decode().split("\n")
    for line in lines:
        if line.startswith(" * nocheck:expect:"):
            checker.add_expected_failure(line[19:])
    checker.test_file(fn)

    # any unexpected failures
    for failure in checker.failures:
        if not failure.expected:
            print(f"{failure.fn} did not expect to see: {failure.message}")
            rc += 1

    # check we got the ones we wanted
    for expect in checker._expected_failure_prefixes:
        if not checker.has_failure(expect):
            print(f"{fn} expected to see: {expect}")
            rc += 1

    # print what we did get
    if rc:
        for failure in checker.failures:
            line_tmp: str = ""
            if failure.fn:
                line_tmp += failure.fn
            if failure.linecnt:
                line_tmp += f":{failure.linecnt}"
            line_tmp += f": {failure.message}"
            print(line_tmp)

    return rc


if __name__ == "__main__":

    parser = argparse.ArgumentParser(description="Check source files")
    parser.add_argument("--test", type=str, help="Run self tests")
    parser.add_argument("--verbose", action="store_true", help="Run in verbose mode")
    args, argv = parser.parse_known_args()

    if args.test:
        _rc: int = 0
        for _fn in glob.glob(f"{args.test}/*.[c|h]"):
            _rc += unit_test(_fn, args.verbose)
        sys.exit(_rc)

    # all done!
    sys.exit(test_files(argv, args.verbose))



#!/usr/bin/env python3
# SPDX-License-Identifier: LGPL-2.1-or-later
#
# Copyright 2020 Canonical Ltd
#
# Authors:
#       Marco Trevisan <marco.trevisan@canonical.com>

import argparse
import importlib.util
import inspect
import os
import unittest


def list_tests(module):
    tests = []
    for name, obj in inspect.getmembers(module):
        if inspect.isclass(obj) and issubclass(obj, unittest.TestCase):
            cases = unittest.defaultTestLoader.getTestCaseNames(obj)
            tests += [(obj, f"{name}.{t}") for t in cases]
    return tests


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("unittest_source", type=argparse.FileType("r"))

    args = parser.parse_args()
    source_path = args.unittest_source.name
    spec = importlib.util.spec_from_file_location(
        os.path.basename(source_path), source_path
    )
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)

    for machine, human in list_tests(module):
        print(human)



#!/usr/bin/env python3
#
# Copyright 2024 Mario Limonciello <mario.limonciello@amd.com>
#
# SPDX-License-Identifier: LGPL-2.1-or-later

import gi
import os
import sys
import unittest
from fwupd_test import FwupdTest

try:
    gi.require_version("Fwupd", "2.0")
    from gi.repository import Fwupd  # pylint: disable=wrong-import-position
except ValueError:
    # when called from unittest-inspector this might not pass, we'll fail later
    # anyway in actual use
    pass


class PciPspNoHsi(FwupdTest):
    """A system that doesn't support exporting any HSI attributes or versions"""

    def setUp(self):
        super().setUp()
        self.testbed.add_from_string(
            """P: /devices/pci0000:20/0000:20:08.1/0000:23:00.1
E: DRIVER=ccp
E: PCI_CLASS=108000
E: PCI_ID=1022:1486
E: PCI_SUBSYS_ID=17AA:1046
E: PCI_SLOT_NAME=0000:23:00.1
E: MODALIAS=pci:v00001022d00001486sv000017AAsd00001046bc10sc80i00
E: SUBSYSTEM=pci
E: ID_PCI_CLASS_FROM_DATABASE=Encryption controller
E: ID_PCI_SUBCLASS_FROM_DATABASE=Encryption controller
E: ID_VENDOR_FROM_DATABASE=Advanced Micro Devices, Inc. [AMD]
E: ID_MODEL_FROM_DATABASE=Starship/Matisse Cryptographic Coprocessor PSPCPP
A: aer_dev_correctable=RxErr 0BadTLP 0BadDLLP 0Rollover 0Timeout 0NonFatalErr 0CorrIntErr 0HeaderOF 0TOTAL_ERR_COR 0
A: aer_dev_fatal=Undefined 0DLP 0SDES 0TLP 0FCP 0CmpltTO 0CmpltAbrt 0UnxCmplt 0RxOF 0MalfTLP 0ECRC 0UnsupReq 0ACSViol 0UncorrIntErr 0BlockedTLP 0AtomicOpBlocked 0TLPBlockedErr 0PoisonTLPBlocked 0TOTAL_ERR_FATAL 0
A: aer_dev_nonfatal=Undefined 0DLP 0SDES 0TLP 0FCP 0CmpltTO 0CmpltAbrt 0UnxCmplt 0RxOF 0MalfTLP 0ECRC 0UnsupReq 0ACSViol 0UncorrIntErr 0BlockedTLP 0AtomicOpBlocked 0TLPBlockedErr 0PoisonTLPBlocked 0TOTAL_ERR_NONFATAL 0
A: ari_enabled=0
A: broken_parity_status=0
A: class=0x108000
H: config=2210861406041000000080101000800000000000000000000000D0C300000000000000000000E0C300000000AA174610000000004800000000000000FF010000
A: consistent_dma_mask_bits=48
A: current_link_speed=16.0 GT/s PCIe
A: current_link_width=16
A: d3cold_allowed=1
A: device=0x1486
A: dma_mask_bits=48
L: driver=../../../../bus/pci/drivers/ccp
A: driver_override=(null)
A: enable=1
L: firmware_node=../../../LNXSYSTM:00/LNXSYBUS:00/PNP0A08:02/device:8f/device:91
L: iommu=../../0000:20:00.2/iommu/ivhd2
L: iommu_group=../../../../kernel/iommu_groups/37
A: irq=256
A: link/l0s_aspm=0
A: link/l1_aspm=0
A: local_cpulist=0-23
A: local_cpus=00000000,00000000,00000000,00ffffff
A: max_link_speed=16.0 GT/s PCIe
A: max_link_width=16
A: modalias=pci:v00001022d00001486sv000017AAsd00001046bc10sc80i00
A: msi_bus=1
A: msi_irqs/257=msix
A: msi_irqs/258=msix
A: numa_node=-1
A: pools=poolinfo - 0.1ccp-1_q4            0    0   64  0ccp-1_q3            0    0   64  0ccp-1_q2            0    0   64  0
A: power/control=on
A: power/runtime_active_time=2767353428
A: power/runtime_status=active
A: power/runtime_suspended_time=0
A: power/wakeup=disabled
A: power/wakeup_abort_count=
A: power/wakeup_active=
A: power/wakeup_active_count=
A: power/wakeup_count=
A: power/wakeup_expire_count=
A: power/wakeup_last_time_ms=
A: power/wakeup_max_time_ms=
A: power/wakeup_total_time_ms=
A: power_state=D0
A: reset_method=flr
A: resource=0x0000000000000000 0x0000000000000000 0x00000000000000000x0000000000000000 0x0000000000000000 0x00000000000000000x00000000c3d00000 0x00000000c3dfffff 0x00000000000402000x0000000000000000 0x0000000000000000 0x00000000000000000x0000000000000000 0x0000000000000000 0x00000000000000000x00000000c3e00000 0x00000000c3e01fff 0x00000000000402000x0000000000000000 0x0000000000000000 0x00000000000000000x0000000000000000 0x0000000000000000 0x00000000000000000x0000000000000000 0x0000000000000000 0x00000000000000000x0000000000000000 0x0000000000000000 0x00000000000000000x0000000000000000 0x0000000000000000 0x00000000000000000x0000000000000000 0x0000000000000000 0x00000000000000000x0000000000000000 0x0000000000000000 0x0000000000000000
A: revision=0x00
A: subsystem_device=0x1046
A: subsystem_vendor=0x17aa
A: vendor=0x1022
"""
        )
        self.testbed.add_from_string(
            """P: /devices/pci0000:20/0000:20:08.1
E: DRIVER=pcieport
E: PCI_CLASS=60400
E: PCI_ID=1022:1484
E: PCI_SUBSYS_ID=1046:17AA
E: PCI_SLOT_NAME=0000:20:08.1
E: MODALIAS=pci:v00001022d00001484sv00001046sd000017AAbc06sc04i00
E: SUBSYSTEM=pci
E: ID_PCI_CLASS_FROM_DATABASE=Bridge
E: ID_PCI_SUBCLASS_FROM_DATABASE=PCI bridge
E: ID_PCI_INTERFACE_FROM_DATABASE=Normal decode
E: ID_VENDOR_FROM_DATABASE=Advanced Micro Devices, Inc. [AMD]
E: ID_MODEL_FROM_DATABASE=Starship/Matisse Internal PCIe GPP Bridge 0 to bus[E:B]
A: aer_dev_correctable=RxErr 0BadTLP 0BadDLLP 0Rollover 0Timeout 0NonFatalErr 0CorrIntErr 0HeaderOF 0TOTAL_ERR_COR 0
A: aer_dev_fatal=Undefined 0DLP 0SDES 0TLP 0FCP 0CmpltTO 0CmpltAbrt 0UnxCmplt 0RxOF 0MalfTLP 0ECRC 0UnsupReq 0ACSViol 0UncorrIntErr 0BlockedTLP 0AtomicOpBlocked 0TLPBlockedErr 0PoisonTLPBlocked 0TOTAL_ERR_FATAL 0
A: aer_dev_nonfatal=Undefined 0DLP 0SDES 0TLP 0FCP 0CmpltTO 0CmpltAbrt 0UnxCmplt 0RxOF 0MalfTLP 0ECRC 0UnsupReq 0ACSViol 0UncorrIntErr 0BlockedTLP 0AtomicOpBlocked 0TLPBlockedErr 0PoisonTLPBlocked 0TOTAL_ERR_NONFATAL 0
A: aer_rootport_total_err_cor=0
A: aer_rootport_total_err_fatal=0
A: aer_rootport_total_err_nonfatal=0
A: ari_enabled=0
A: broken_parity_status=0
A: class=0x060400
H: config=22108414070410000000040610000100000000000000000020232300F1010000C0C3E0C3F1FF01000000000000000000000000005000000000000000FF011200
A: consistent_dma_mask_bits=32
A: current_link_speed=16.0 GT/s PCIe
A: current_link_width=16
A: d3cold_allowed=1
A: device=0x1484
A: dma_mask_bits=32
L: driver=../../../bus/pci/drivers/pcieport
A: driver_override=(null)
A: enable=2
L: firmware_node=../../LNXSYSTM:00/LNXSYBUS:00/PNP0A08:02/device:8f
L: iommu=../0000:20:00.2/iommu/ivhd2
L: iommu_group=../../../kernel/iommu_groups/33
A: irq=43
A: local_cpulist=0-23
A: local_cpus=00000000,00000000,00000000,00ffffff
A: max_link_speed=16.0 GT/s PCIe
A: max_link_width=16
A: modalias=pci:v00001022d00001484sv00001046sd000017AAbc06sc04i00
A: msi_bus=1
A: msi_irqs/43=msi
A: numa_node=-1
A: power/autosuspend_delay_ms=100
A: power/control=auto
A: power/runtime_active_time=2767353433
A: power/runtime_status=active
A: power/runtime_suspended_time=0
A: power/wakeup=enabled
A: power/wakeup_abort_count=0
A: power/wakeup_active=0
A: power/wakeup_active_count=0
A: power/wakeup_count=0
A: power/wakeup_expire_count=0
A: power/wakeup_last_time_ms=0
A: power/wakeup_max_time_ms=0
A: power/wakeup_total_time_ms=0
A: power_state=D0
A: reset_method=pm
A: resource=0x0000000000000000 0x0000000000000000 0x00000000000000000x0000000000000000 0x0000000000000000 0x00000000000000000x0000000000000000 0x0000000000000000 0x00000000000000000x0000000000000000 0x0000000000000000 0x00000000000000000x0000000000000000 0x0000000000000000 0x00000000000000000x0000000000000000 0x0000000000000000 0x00000000000000000x0000000000000000 0x0000000000000000 0x00000000000000000x0000000000000000 0x0000000000000000 0x00000000000000000x0000000000000000 0x0000000000000000 0x00000000000000000x0000000000000000 0x0000000000000000 0x00000000000000000x0000000000000000 0x0000000000000000 0x00000000000000000x0000000000000000 0x0000000000000000 0x00000000000000000x0000000000000000 0x0000000000000000 0x00000000000000000x0000000000000000 0x0000000000000000 0x00000000000000000x00000000c3c00000 0x00000000c3efffff 0x00000000000002000x0000000000000000 0x0000000000000000 0x00000000000000000x0000000000000000 0x0000000000000000 0x0000000000000000
A: revision=0x00
A: secondary_bus_number=35
A: subordinate_bus_number=35
A: subsystem_device=0x17aa
A: subsystem_vendor=0x1046
A: vendor=0x1022
"""
        )

    def test_pci_psp_device(self):
        """Verify the PCI PSP device is detected correctly"""
        self.start_daemon()
        devices = Fwupd.Client().get_devices()

        count = 0
        for dev in devices:
            if dev.get_plugin() != "pci_psp":
                continue
            self.assertEqual(dev.get_name(), "Secure Processor")
            self.assertEqual(dev.get_version(), None)
            self.assertEqual(dev.get_version_bootloader(), None)
            count += 1
        self.assertEqual(count, 1)

    def test_pci_psp_hsi(self):
        """Verify the PCI PSP HSI attributes are detected correctly"""
        self.start_daemon()
        attrs = Fwupd.Client().get_host_security_attrs()
        for attr in attrs:
            if attr.get_plugin() != "pci_psp":
                continue
            if attr.get_appstream_id() == "org.fwupd.hsi.PlatformFused":
                self.assertEqual(attr.get_name(), "Fused platform")
                self.assertEqual(attr.get_result(), Fwupd.SecurityAttrResult.UNKNOWN)
                self.assertEqual(attr.get_level(), 1)
            elif attr.get_appstream_id() == "org.fwupd.hsi.EncryptedRam":
                self.assertEqual(attr.get_name(), "Encrypted RAM")
                self.assertEqual(attr.get_result(), Fwupd.SecurityAttrResult.UNKNOWN)
                self.assertEqual(attr.get_level(), 4)
            elif attr.get_appstream_id() == "org.fwupd.hsi.Amd.RollbackProtection":
                self.assertEqual(attr.get_name(), "Processor rollback protection")
                self.assertEqual(attr.get_result(), Fwupd.SecurityAttrResult.UNKNOWN)
                self.assertEqual(attr.get_level(), 4)
            elif attr.get_appstream_id() == "org.fwupd.hsi.Amd.SpiReplayProtection":
                self.assertEqual(attr.get_name(), "SPI replay protection")
                self.assertEqual(attr.get_result(), Fwupd.SecurityAttrResult.UNKNOWN)
                self.assertEqual(attr.get_level(), 3)
            elif attr.get_appstream_id() == "org.fwupd.hsi.PlatformDebugLocked":
                self.assertEqual(attr.get_name(), "Platform debugging")
                self.assertEqual(attr.get_result(), Fwupd.SecurityAttrResult.UNKNOWN)
                self.assertEqual(attr.get_level(), 2)
            elif attr.get_appstream_id() == "org.fwupd.hsi.Amd.SpiWriteProtection":
                self.assertEqual(attr.get_name(), "SPI write protection")
                self.assertEqual(attr.get_result(), Fwupd.SecurityAttrResult.UNKNOWN)
                self.assertEqual(attr.get_level(), 2)


class PciPspWithHsi(FwupdTest):
    """A system that can export both HSI attributes and versions"""

    def setUp(self):
        super().setUp()
        self.testbed.add_from_string(
            """P: /devices/pci0000:00/0000:00:08.1/0000:c1:00.2
E: DRIVER=ccp
E: ID_MODEL_FROM_DATABASE=Family 19h (Model 74h) CCP/PSP 3.0 Device
E: ID_PATH=pci-0000:c1:00.2
E: ID_PATH_TAG=pci-0000_c1_00_2
E: ID_PCI_CLASS_FROM_DATABASE=Encryption controller
E: ID_PCI_SUBCLASS_FROM_DATABASE=Encryption controller
E: ID_VENDOR_FROM_DATABASE=Advanced Micro Devices, Inc. [AMD]
E: MODALIAS=pci:v00001022d000015C7sv0000F111sd00000006bc10sc80i00
E: PCI_CLASS=108000
E: PCI_ID=1022:15C7
E: PCI_SLOT_NAME=0000:c1:00.2
E: PCI_SUBSYS_ID=F111:0006
E: SUBSYSTEM=pci
A: anti_rollback_status=0
A: ari_enabled=0
A: bootloader_version=00.2d.00.78
A: broken_parity_status=0
A: class=0x108000
H: config=2210C715070410000000801010008000000000000000000000004090000000000000000000C05C900000000011F10600000000004800000000000000FF030000
A: consistent_dma_mask_bits=48
A: current_link_speed=16.0 GT/s PCIe
A: current_link_width=16
A: d3cold_allowed=1
A: debug_lock_on=1
A: device=0x15c7
A: dma_mask_bits=48
L: driver=../../../../bus/pci/drivers/ccp
A: driver_override=(null)
A: enable=1
L: firmware_node=../../../LNXSYSTM:00/LNXSYBUS:00/PNP0A08:00/device:16/device:18
A: fused_part=1
A: hsp_tpm_available=0
L: iommu=../../0000:00:00.2/iommu/ivhd0
L: iommu_group=../../../../kernel/iommu_groups/16
A: irq=109
A: link/l0s_aspm=0
A: link/l1_aspm=0
A: local_cpulist=0-11
A: local_cpus=0fff
A: max_link_speed=16.0 GT/s PCIe
A: max_link_width=16
A: modalias=pci:v00001022d000015C7sv0000F111sd00000006bc10sc80i00
A: msi_bus=1
A: msi_irqs/110=msix
A: msi_irqs/111=msix
A: numa_node=-1
A: power/control=on
A: power/runtime_active_time=17568317
A: power/runtime_status=active
A: power/runtime_suspended_time=0
A: power/wakeup=disabled
A: power/wakeup_abort_count=
A: power/wakeup_active=
A: power/wakeup_active_count=
A: power/wakeup_count=
A: power/wakeup_expire_count=
A: power/wakeup_last_time_ms=
A: power/wakeup_max_time_ms=
A: power/wakeup_total_time_ms=
A: power_state=D0
A: resource=0x0000000000000000 0x0000000000000000 0x00000000000000000x0000000000000000 0x0000000000000000 0x00000000000000000x0000000090400000 0x00000000904fffff 0x00000000000402000x0000000000000000 0x0000000000000000 0x00000000000000000x0000000000000000 0x0000000000000000 0x00000000000000000x00000000905cc000 0x00000000905cdfff 0x00000000000402000x0000000000000000 0x0000000000000000 0x00000000000000000x0000000000000000 0x0000000000000000 0x00000000000000000x0000000000000000 0x0000000000000000 0x00000000000000000x0000000000000000 0x0000000000000000 0x00000000000000000x0000000000000000 0x0000000000000000 0x00000000000000000x0000000000000000 0x0000000000000000 0x00000000000000000x0000000000000000 0x0000000000000000 0x0000000000000000
A: revision=0x00
A: rom_armor_enforced=1
A: rpmc_production_enabled=0
A: rpmc_spirom_available=1
A: subsystem_device=0x0006
A: subsystem_vendor=0xf111
A: tee_version=00.2d.00.78
A: tsme_status=0
A: vendor=0x1022
"""
        )
        self.testbed.add_from_string(
            """P: /devices/pci0000:00/0000:00:08.1
E: DRIVER=pcieport
E: ID_PATH=pci-0000:00:08.1
E: ID_PATH_TAG=pci-0000_00_08_1
E: ID_PCI_CLASS_FROM_DATABASE=Bridge
E: ID_PCI_INTERFACE_FROM_DATABASE=Normal decode
E: ID_PCI_SUBCLASS_FROM_DATABASE=PCI bridge
E: ID_VENDOR_FROM_DATABASE=Advanced Micro Devices, Inc. [AMD]
E: MODALIAS=pci:v00001022d000014EBsv00000006sd0000F111bc06sc04i00
E: PCI_CLASS=60400
E: PCI_ID=1022:14EB
E: PCI_SLOT_NAME=0000:00:08.1
E: PCI_SUBSYS_ID=0006:F111
E: SUBSYSTEM=pci
A: ari_enabled=0
A: broken_parity_status=0
A: class=0x060400
H: config=2210EB14070410000000040610008100000000000000000000C1C1001111000000905090010071107800000078000000000000005000000000000000FF010200
A: consistent_dma_mask_bits=32
A: current_link_speed=16.0 GT/s PCIe
A: current_link_width=16
A: d3cold_allowed=1
A: device=0x14eb
A: dma_mask_bits=32
L: driver=../../../bus/pci/drivers/pcieport
A: driver_override=(null)
A: enable=2
L: firmware_node=../../LNXSYSTM:00/LNXSYBUS:00/PNP0A08:00/device:16
L: iommu=../0000:00:00.2/iommu/ivhd0
L: iommu_group=../../../kernel/iommu_groups/7
A: irq=42
A: local_cpulist=0-11
A: local_cpus=0fff
A: max_link_speed=16.0 GT/s PCIe
A: max_link_width=16
A: modalias=pci:v00001022d000014EBsv00000006sd0000F111bc06sc04i00
A: msi_bus=1
A: msi_irqs/42=msi
A: numa_node=-1
A: power/autosuspend_delay_ms=100
A: power/control=auto
A: power/runtime_active_time=17568337
A: power/runtime_status=active
A: power/runtime_suspended_time=0
A: power/wakeup=disabled
A: power/wakeup_abort_count=
A: power/wakeup_active=
A: power/wakeup_active_count=
A: power/wakeup_count=
A: power/wakeup_expire_count=
A: power/wakeup_last_time_ms=
A: power/wakeup_max_time_ms=
A: power/wakeup_total_time_ms=
A: power_state=D0
A: reset_method=pm
A: resource=0x0000000000000000 0x0000000000000000 0x00000000000000000x0000000000000000 0x0000000000000000 0x00000000000000000x0000000000000000 0x0000000000000000 0x00000000000000000x0000000000000000 0x0000000000000000 0x00000000000000000x0000000000000000 0x0000000000000000 0x00000000000000000x0000000000000000 0x0000000000000000 0x00000000000000000x0000000000000000 0x0000000000000000 0x00000000000000000x0000000000000000 0x0000000000000000 0x00000000000000000x0000000000000000 0x0000000000000000 0x00000000000000000x0000000000000000 0x0000000000000000 0x00000000000000000x0000000000000000 0x0000000000000000 0x00000000000000000x0000000000000000 0x0000000000000000 0x00000000000000000x0000000000000000 0x0000000000000000 0x00000000000000000x0000000000001000 0x0000000000001fff 0x00000000000001010x0000000090000000 0x00000000905fffff 0x00000000000002000x0000007800000000 0x00000078107fffff 0x00000000001022010x0000000000000000 0x0000000000000000 0x0000000000000000
A: revision=0x00
A: secondary_bus_number=193
A: subordinate_bus_number=193
A: subsystem_device=0xf111
A: subsystem_vendor=0x0006
A: vendor=0x1022
"""
        )

    def test_pci_psp_device(self):
        """Verify the PCI PSP device is detected correctly"""
        self.start_daemon()
        devices = Fwupd.Client().get_devices()

        count = 0
        for dev in devices:
            if dev.get_plugin() != "pci_psp":
                continue
            self.assertEqual(dev.get_name(), "Secure Processor")
            self.assertEqual(dev.get_version(), "00.2d.00.78")
            self.assertEqual(dev.get_version_bootloader(), "00.2d.00.78")
            count += 1
        self.assertEqual(count, 1)

    def test_pci_psp_hsi(self):
        """Verify the PCI PSP HSI attributes are detected correctly"""
        self.start_daemon()
        attrs = Fwupd.Client().get_host_security_attrs()
        for attr in attrs:
            if attr.get_plugin() != "pci_psp":
                continue
            if attr.get_appstream_id() == "org.fwupd.hsi.PlatformFused":
                self.assertEqual(attr.get_name(), "Fused platform")
                self.assertEqual(attr.get_result(), Fwupd.SecurityAttrResult.LOCKED)
                self.assertEqual(attr.get_level(), 1)
            elif attr.get_appstream_id() == "org.fwupd.hsi.EncryptedRam":
                self.assertEqual(attr.get_name(), "Encrypted RAM")
                self.assertEqual(
                    attr.get_result(), Fwupd.SecurityAttrResult.NOT_SUPPORTED
                )
                self.assertEqual(attr.get_level(), 4)
            elif attr.get_appstream_id() == "org.fwupd.hsi.Amd.RollbackProtection":
                self.assertEqual(attr.get_name(), "Processor rollback protection")
                self.assertEqual(
                    attr.get_result(), Fwupd.SecurityAttrResult.NOT_ENABLED
                )
                self.assertEqual(attr.get_level(), 4)
            elif attr.get_appstream_id() == "org.fwupd.hsi.Amd.SpiReplayProtection":
                self.assertEqual(attr.get_name(), "SPI replay protection")
                self.assertEqual(
                    attr.get_result(), Fwupd.SecurityAttrResult.NOT_ENABLED
                )
                self.assertEqual(attr.get_level(), 3)
            elif attr.get_appstream_id() == "org.fwupd.hsi.PlatformDebugLocked":
                self.assertEqual(attr.get_name(), "Platform debugging")
                self.assertEqual(attr.get_result(), Fwupd.SecurityAttrResult.LOCKED)
                self.assertEqual(attr.get_level(), 2)
            elif attr.get_appstream_id() == "org.fwupd.hsi.Amd.SpiWriteProtection":
                self.assertEqual(attr.get_name(), "SPI write protection")
                self.assertEqual(attr.get_result(), Fwupd.SecurityAttrResult.ENABLED)
                self.assertEqual(attr.get_level(), 2)


if __name__ == "__main__":
    # run ourselves under umockdev
    if "umockdev" not in os.environ.get("LD_PRELOAD", ""):
        os.execvp("umockdev-wrapper", ["umockdev-wrapper", sys.executable] + sys.argv)

    prog = unittest.main(exit=False)
    if prog.result.errors or prog.result.failures:
        sys.exit(1)

    # Translate to skip error
    if prog.result.testsRun == len(prog.result.skipped):
        sys.exit(77)



#!/usr/bin/env python3
#
# Copyright 2026 Mario Limonciello <superm1@gmail.com>
#
# SPDX-License-Identifier: LGPL-2.1-or-later

import os
import sys


def test_files() -> int:
    rc: int = 0
    hsi_dir = "docs/hsi-tests.d"
    meson_build = os.path.join(hsi_dir, "meson.build")

    with open(meson_build) as f:
        meson_contents = f.read()

    for fn in sorted(os.listdir(hsi_dir)):
        if fn == "meson.build":
            continue
        if not fn.endswith(".json"):
            print(f"{hsi_dir}/{fn} does not have a .json extension")
            rc = 1
            continue
        if f"'{fn}'" not in meson_contents:
            print(f"{hsi_dir}/{fn} is missing from {meson_build}")
            rc = 1

    return rc


if __name__ == "__main__":
    # all done!
    sys.exit(test_files())



#!/usr/bin/env python3
# Copyright 2025 Richard Hughes <richard@hughsie.com>
#
# SPDX-License-Identifier: LGPL-2.1-or-later
#
# pylint: disable=invalid-name,missing-module-docstring

import sys
import struct
import uuid

buf = b""
buf += struct.pack("<4s", b"UEFI")  # signature
buf += struct.pack("<I", 70)  # length of buf
buf += struct.pack("<B", 1)  # revision
buf += struct.pack("<B", 0x65)  # checksum
buf += struct.pack("<6s", b"XXXXXX")  # oem_id
buf += struct.pack("<8s", b"YYYYYYYY")  # oem_table_id
buf += struct.pack("<I", 1)  # oem_revision
buf += struct.pack("<4s", b"FWPD")  # asl_compiler_id
buf += struct.pack("<I", 1)  # asl_compiler_revision

# then GUID @0x24
buf += uuid.UUID("9d4bf935-a674-4710-ba02-bf0aa1758c7b").bytes_le

# then jumk data
buf += struct.pack("<4s", b"JUNK")

# add insyde blob
buf += struct.pack("<6s", b"$QUIRK")  # signature
buf += struct.pack("<I", 14)  # size
buf += struct.pack("<I", 1)  # flags, COD does not work

with open(sys.argv[1], "wb") as f:
    f.write(buf)



#!/usr/bin/env python3
#
# Copyright 2025 Richard Hughes <richard@hughsie.com>
#
# SPDX-License-Identifier: GPL-2.0-or-later
#
# pylint: disable=missing-module-docstring,missing-class-docstring,
# pylint: disable=missing-function-docstring,too-few-public-methods,consider-using-enumerate

from typing import Optional, List
from enum import Enum
from fnmatch import fnmatch
import sys
import copy


class TokenHint(Enum):

    COMMENT = 1
    STRING = 2
    DEFINE = 3
    ENUM = 4
    LBRACKET = 5
    RBRACKET = 6
    FUNCTION = 7
    MACRO = 8
    INTEGER = 9
    STRUCT = 10

    @classmethod
    def value_of(cls, value):
        for k, v in cls.__members__.items():
            if k == value:
                return v
        raise ValueError(f"'{cls.__name__}' enum not found for '{value}'")


def _check_int(s: str) -> bool:
    if s.startswith("-"):
        s = s[1:]
    if s.startswith("0x"):
        s = s[2:]
    return s.isdigit()


class Token:

    def __init__(self, data: str, linecnt: int = 0, hint: Optional[TokenHint] = None):

        self.linecnt: int = linecnt
        self.linecnt_end: int = linecnt
        self.data: str = data
        self.hint: Optional[TokenHint] = hint

        # autohint
        if not self.hint:
            if data.startswith('"') and data.endswith('"'):
                self.hint = TokenHint.STRING
            elif data.startswith("/*") and data.endswith("*/"):
                self.hint = TokenHint.COMMENT
            elif data == "(":
                self.hint = TokenHint.LBRACKET
            elif data == ")":
                self.hint = TokenHint.RBRACKET
            elif _check_int(data):
                self.hint = TokenHint.INTEGER

    def __repr__(self) -> str:

        tmp = [f"linecnt={self.linecnt}"]
        if self.linecnt_end != self.linecnt:
            tmp.append(f"linecnt_end={self.linecnt_end}")
        if self.data:
            tmp.append(f"data='{self.data}'")
        if self.hint:
            tmp.append(f"hint={self.hint}")
        return f"Token({', '.join(tmp)})"


def _token_fuzzy_match(token: Token, data: str) -> bool:
    """data is of the form:
    * `literal`
    * `~fnmatch`
    * `@HINT`
    * `literal@HINT`
    * `~fnmatch@HINT`
    """

    # hint
    query: List[str] = data.split("@")
    try:
        if token.hint != TokenHint.value_of(query[1]):
            return False
    except IndexError:
        pass

    # fnmatch
    if query[0].startswith("~"):
        query_fnmatch = query[0][1:]
        if not query_fnmatch or query_fnmatch == "*":
            return True
        return fnmatch(token.data, query_fnmatch)

    # literal
    if query[0]:
        return token.data == query[0]

    # empty
    return True


class TokenList(list):

    def __init__(self, tokens: Optional[list[Token]] = None):

        for token in tokens or []:
            self.append(token)

    def append(self, token: Token) -> None:

        # autohint
        if not token.hint and token.data != ";":
            if len(self) >= 1 and self[-1].data == "#define":
                token.hint = TokenHint.DEFINE
            elif len(self) >= 1 and self[-1].data == "enum":
                token.hint = TokenHint.ENUM
            elif len(self) >= 1 and self[-1].data == "struct":
                token.hint = TokenHint.STRUCT

        # autohint previous token
        if len(self) > 0 and token.hint == TokenHint.LBRACKET:
            token_prev: Token = self[-1]
            if not token_prev.hint:
                if token_prev.data.find("_") != -1:
                    if token_prev.data.upper() == token_prev.data:
                        token_prev.hint = TokenHint.MACRO
                    else:
                        token_prev.hint = TokenHint.FUNCTION

        # add
        list.append(self, token)

    def find_fuzzy(
        self,
        data_fuzzy: List[str],
        offset: Optional[int] = None,
        reverse: bool = False,
        skip_comments: bool = False,
    ) -> int:
        """
        Look for a fuzzy token sequence and return the position to the
        first token. Returns -1 if not found.
        """
        if reverse:
            pos_range = range(offset or len(self) - 1, 0, -1)
        else:
            pos_range = range(offset or 0, len(self) - (len(data_fuzzy) - 1))
        for pos in pos_range:

            all_match: bool = True
            for datapos in range(len(data_fuzzy)):
                comment_offset: int = 0
                if skip_comments and self[pos + datapos].hint == TokenHint.COMMENT:
                    comment_offset = 1
                token = self[pos + datapos + comment_offset]
                if not _token_fuzzy_match(token, data_fuzzy[datapos]):
                    all_match = False
                    break
            if all_match:
                return pos
        return -1

    def endswith_fuzzy(self, data_fuzzy: List[str]) -> bool:
        """
        Look for a fuzzy token sequence at the end of the list.
        Returns False if not found.
        """
        offset = len(self) - len(data_fuzzy)
        if offset < 0:
            return False
        for datapos in range(len(data_fuzzy)):
            token = self[offset + datapos]
            if not _token_fuzzy_match(token, data_fuzzy[datapos]):
                return False
        return True

    def count_fuzzy(self, data_fuzzy: List[str]) -> int:
        """
        Return the number of fuzzy matches matching all tokens.
        """
        cnt: int = 0
        for pos in range(len(self) - (len(data_fuzzy) - 1)):
            all_match: bool = True
            for datapos in range(len(data_fuzzy)):
                if not _token_fuzzy_match(self[pos + datapos], data_fuzzy[datapos]):
                    all_match = False
                    break
            if all_match:
                cnt += 1
        return cnt


class NodeHint(Enum):

    UNION = 1
    ENUM = 2
    STRUCT = 3
    STRUCT_TYPEDEF = 4
    STRUCT_CONST = 5


class Node:

    def __init__(
        self, depth: int, linecnt: int, tokens_pre: Optional[TokenList] = None
    ):
        self.depth: int = depth
        self.linecnt: int = linecnt
        self.linecnt_end: int = linecnt
        self.tokens_pre: TokenList = tokens_pre or TokenList()
        self.tokens: TokenList = TokenList()
        self.hint: Optional[NodeHint] = None

    def __repr__(self) -> str:

        tmp = [f"depth={self.depth}", f"linecnt={self.linecnt}"]
        if self.linecnt_end != self.linecnt:
            tmp.append(f"linecnt_end={self.linecnt_end}")
        if self.hint:
            tmp.append(f"hint={self.hint}")
        if self.tokens_pre:
            tmp.append(f"tokens_pre={[token.data for token in self.tokens_pre]}")
        if self.tokens:
            tmp.append(f"tokens={[token.data for token in self.tokens]}")
        return f"Node({', '.join(tmp)})"


class Tokenizer:
    def __init__(self, data: str):
        self.tokens: TokenList = TokenList()
        self._nodes: List[Node] = []
        self._acc: str = ""
        self._linecnt: int = 1
        if data:
            self._parse(data)

    def _add_token(self, token: Token) -> None:

        # merge into previous token
        if self.tokens:
            token_prev: Token = self.tokens[-1]

            # STRING -> STRING
            if token_prev.hint == TokenHint.STRING and token.hint == TokenHint.STRING:
                new = token_prev.data[:-1] + token.data[1:]
                token_prev.data = new
                token_prev.linecnt_end = token.linecnt
                return

            # COMMENT -> COMMENT
            if token_prev.hint == TokenHint.COMMENT and token.hint == TokenHint.COMMENT:
                new = token_prev.data + " " + token.data
                token_prev.data = new
                token_prev.linecnt_end = token.linecnt
                return

            # "-" -> INTEGER
            if token_prev.data == "-" and token.hint == TokenHint.INTEGER:
                new = token_prev.data + token.data
                token_prev.data = new
                return

        # add new
        self.tokens.append(token)

    def dump_acc(self, hint: Optional[TokenHint] = None) -> None:

        stripped = self._acc.strip()
        if stripped:
            if fnmatch(stripped, "G_G*_FORMAT"):
                stripped = '"%u"'
            self._add_token(Token(stripped, linecnt=self._linecnt, hint=hint))
        self._acc = ""

    def _parse(self, data: str):

        is_quote_double: bool = False
        is_quote_single: bool = False
        is_comment_mode: bool = False
        is_escape_mode: bool = False

        for pos, char in enumerate(data):

            # newline
            if char == "\n":

                if is_comment_mode:
                    self.dump_acc(hint=TokenHint.COMMENT)
                else:
                    self.dump_acc()
                self._linecnt += 1
                continue

            if char == "\\" and not is_escape_mode:
                is_escape_mode = True
                continue

            # string quotes
            if (
                char == '"'
                and not is_escape_mode
                and not is_quote_single
                and not is_comment_mode
            ):
                is_quote_double = not is_quote_double
            if (
                char == "'"
                and not is_escape_mode
                and not is_quote_double
                and not is_comment_mode
            ):
                is_quote_single = not is_quote_single

            # delimiter
            dump_char: bool = False
            if (
                char
                in [
                    " ",
                    "*",
                    "(",
                    ")",
                    "{",
                    "}",
                    ",",
                    "&",
                    ";",
                    ">",
                    "<",
                    "-",
                    "[",
                    "]",
                    "!",
                ]
                and not is_comment_mode
                and not is_quote_double
                and not is_quote_single
            ):
                self.dump_acc()
                dump_char = True

            # comment
            if data[pos : pos + 2] == "/*" and not is_quote_double:
                is_comment_mode = True

            if is_escape_mode:
                self._acc += "\\"
            self._acc += char
            is_escape_mode = False

            # end comment
            if data[pos - 1 : pos + 1] == "*/" and not is_quote_double:
                self.dump_acc(hint=TokenHint.COMMENT)
                is_comment_mode = False

            # delimiter
            if dump_char:
                self.dump_acc()

        # any left
        self.dump_acc()

    def _add_node(self, node: Node):

        # auto-hint node
        if not node.hint:
            try:
                if node.tokens_pre[-1].data == "union":
                    node.hint = NodeHint.UNION
                elif node.tokens_pre[-1].data == "enum":
                    node.hint = NodeHint.ENUM
                elif node.tokens_pre[-1].data == "struct":
                    node.hint = NodeHint.STRUCT
                    if node.tokens_pre[-2].data == "typedef":
                        node.hint = NodeHint.STRUCT_TYPEDEF
                    elif node.tokens_pre[-2].data == "const":
                        node.hint = NodeHint.STRUCT_CONST
            except IndexError:
                pass

        # add
        self._nodes.append(node)

    def _ensure_nodes(self) -> None:

        tokens_acc: TokenList = TokenList()
        depth: int = 0
        stack: List[Node] = []
        node_parent: Optional[Node] = None

        for token in self.tokens:

            # ignore __attribute__
            if token.data == "G_GNUC_FLAG_ENUM":
                continue

            # start, end or continue
            if token.data == "{":
                if stack:
                    stack[-1].tokens.extend(tokens_acc)
                node = Node(
                    depth=len(stack),
                    linecnt=token.linecnt,
                    tokens_pre=copy.copy(tokens_acc),
                )
                stack.append(node)
                self._add_node(node)
                tokens_acc.clear()
                depth += 1
            elif token.data == "}":
                if not stack:
                    raise ValueError(f"unequal nesting on line {token.linecnt}")
                node_parent = stack.pop()
                node_parent.tokens.extend(tokens_acc)
                node_parent.linecnt_end = token.linecnt
                tokens_acc.clear()
                depth -= 1
            else:
                # reposition typedef struct names to be before the {
                if (
                    node_parent
                    and node_parent.tokens_pre
                    and node_parent.tokens_pre[-1].data
                    in [
                        "enum",
                        "struct",
                        "union",
                    ]
                ):
                    node_parent.tokens_pre.append(token)
                else:
                    tokens_acc.append(token)

        # create a fake node
        if not self._nodes and tokens_acc:
            self._add_node(
                Node(depth=0, linecnt=tokens_acc[0].linecnt, tokens_pre=tokens_acc)
            )

        # sanity check
        if depth != 0:
            raise ValueError("has unequal nesting")

    @property
    def nodes(self) -> List[Node]:

        if not self._nodes:
            self._ensure_nodes()
        return self._nodes


if __name__ == "__main__":

    for fn in sys.argv[1:]:
        with open(fn, "rb") as f:
            tok = Tokenizer(f.read().decode())
            print(tok.tokens)
            print(tok.nodes)



#!/usr/bin/env python3
# pylint: disable=invalid-name,missing-module-docstring,missing-function-docstring
#
# Copyright 2025 Richard Hughes <richard@hughsie.com>
#
# SPDX-License-Identifier: LGPL-2.1-or-later

import subprocess
import sys

if __name__ == "__main__":
    rc: int = 0

    # load the manpage, bash-completion, etc
    data: dict[str, str] = {}
    for fn in [
        "src/fwupdmgr.md",
        "data/bash-completion/fwupdmgr",
        "data/fish-completion/fwupdmgr.fish",
    ]:
        with open(fn, "rb") as f:
            data[fn] = f.read().decode()

    # if we can't run the binary for some reason, assume everything is okay
    try:
        pr = subprocess.run(
            ["venv/build/src/fwupdmgr", "get-actions", "--force"],
            cwd=".",
            # stderr=subprocess.PIPE,
            stdout=subprocess.PIPE,
            encoding="utf-8",
        )
    except (FileNotFoundError, subprocess.CalledProcessError):
        sys.exit(0)

    # check each action is at least mentioned in each file
    for fn, txt in data.items():
        for action in pr.stdout.split("\n"):
            if txt.find(action) == -1:
                print(f"* CLI action {action} not documented in {fn}")
                rc = 1
    sys.exit(rc)
